@@ -32,7 +32,7 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: arg.c,v 1.36 2001/02/13 02:06:23 abe Exp $";
+static char *rcsid = "$Id: arg.c,v 1.48 2008/10/21 16:21:41 abe Exp $";
 #endif
 
 
@@ -58,9 +58,10 @@ static int NCmdRxA = 0;			/* space allocated to CmdRx[] */
  */
 
 _PROTOTYPE(static int ckfd_range,(char *first, char *dash, char *last, int *lo, int *hi));
-_PROTOTYPE(static void enter_fd_lst,(char *nm, int lo, int hi));
+_PROTOTYPE(static int enter_fd_lst,(char *nm, int lo, int hi, int excl));
 _PROTOTYPE(static int enter_nwad,(struct nwad *n, int sp, int ep, char *s, struct hostent *he));
 _PROTOTYPE(static struct hostent *lkup_hostnm,(char *hn, struct nwad *n));
+_PROTOTYPE(static char *isIPv4addr,(char *hn, unsigned char *a, int al));
 
 
 /*
@@ -127,10 +128,9 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 	struct stat *sbp;	/* if non-NULL, pointer to stat(2) buffer
 				 * when argument count == 1 */
 {
-	unsigned char ad, an;
+	char *ap, *fnm, *fsnm, *path;
 	short err = 0;
-	char *fnm, *fsnm, *path;
-	int fsm, ftype, j;
+	int fsm, ftype, j, k;
 	MALLOC_S l;
 	struct mounts *mp;
 	static struct mounts **mmp = (struct mounts **)NULL;
@@ -145,6 +145,7 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 #endif	/* defined(CKFA_EXPDEV) */
 
 #if	defined(HASPROCFS)
+	unsigned char ad, an;
 	int pfsnl = -1;
 	pid_t pid;
 	struct procfsid *pfi;
@@ -163,14 +164,29 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 		}
 	    }
 	/*
-	 * Remove a terminating `/' from Readlink()-expanded path, but never
-	 * from an argument-specified path.  Don't remove `/' from a single
-	 * character path name, either.
-	 *
+	 * Remove terminating `/' characters from paths longer than one.
+	 */
+	    j = k = strlen(path);
+	    while ((k > 1) && (path[k-1] == '/')) {
+		k--;
+	    }
+	    if (k < j) {
+		if (path != av[i])
+		    path[k] = '\0';
+		else {
+		    if (!(ap = (char *)malloc((MALLOC_S)(k + 1)))) {
+			(void) fprintf(stderr, "%s: no space for copy of %s\n",
+			    Pn, path);
+			Exit(1);
+		    }
+		    (void) strncpy(ap, path, k);
+		    ap[k] = '\0';
+		    path = ap;
+		}
+	    }
+	/*
 	 * Check for file system argument.
 	 */
-	    if (path != av[i] && (j = strlen(path)) > 1 && path[j-1] == '/')
-		path[j-1] = '\0';
 	    for (ftype = 1, mp = readmnt(), nm = 0;
 		 (fv != 1) && mp;
 		 mp = mp->next)
@@ -271,7 +287,7 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 #endif	/* defined(HASSPECDEVD) */
 
 		    }
-		    sfp->i = sb.st_ino;
+		    sfp->i = (INODETYPE)sb.st_ino;
 		    sfp->mode = sb.st_mode & S_IFMT;
 		
 #if	defined(CKFA_EXPDEV)
@@ -331,7 +347,11 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 	     */
 		if (!fnm || fnm == path) {
 		    sfp->name = fnm;
+
+#if	defined(HASPROCFS)
 		    an = 0;
+#endif	/* defined(HASPROCFS) */
+
 		} else {
 		    if (!(sfp->name = mkstrcpy(fnm, (MALLOC_S *)NULL))) {
 			(void) fprintf(stderr,
@@ -339,11 +359,19 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 			safestrprt(fnm, stderr, 1);
 			Exit(1);
 		    }
+
+#if	defined(HASPROCFS)
 		    an = 1;
+#endif	/* defined(HASPROCFS) */
+
 		}
 		if (!fsnm || fsnm == path) {
 		    sfp->devnm = fsnm;
+
+#if	defined(HASPROCFS)
 		    ad = 0;
+#endif	/* defined(HASPROCFS) */
+
 		} else {
 		    if (!(sfp->devnm = mkstrcpy(fsnm, (MALLOC_S *)NULL))) {
 			(void) fprintf(stderr,
@@ -351,7 +379,11 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 			safestrprt(fsnm, stderr, 1);
 			Exit(1);
 		    }
+
+#if	defined(HASPROCFS)
 		    ad = 1;
+#endif	/* defined(HASPROCFS) */
+
 		}
 		if (!(sfp->aname = mkstrcpy(av[i], (MALLOC_S *)NULL))) {
 		    (void) fprintf(stderr,
@@ -418,7 +450,7 @@ ck_file_arg(i, ac, av, fv, rs, sbp)
 		Procfsid = pfi;
 
 # if	defined(HASPINODEN)
-		pfi->inode = (unsigned long)sfp->i;
+		pfi->inode = (INODETYPE)sfp->i;
 # endif	/* defined(HASPINODEN) */
 
 	    /*
@@ -523,7 +555,7 @@ ctrl_dcache(c)
 	for (c++; *c && (*c == ' ' || *c == '\t'); c++)
 	    ;
 	if (strlen(c)) {
-	    if (!(DCpathArg = mkstrcpy(c, (MALLOC_S)NULL))) {
+	    if (!(DCpathArg = mkstrcpy(c, (MALLOC_S *)NULL))) {
 		(void) fprintf(stderr, "%s: no space for -D path: ", Pn);
 		safestrprt(c, stderr, 1);
 		Exit(1);
@@ -699,7 +731,7 @@ enter_fd(f)
 	char *f;			/* file descriptor list pointer */
 {
 	char c, *cp1, *cp2, *dash;
-	int err, hi, lo;
+	int err, excl, hi, lo;
 	char *fc;
 /*
  *  Check for non-empty list and make a copy.
@@ -720,8 +752,15 @@ enter_fd(f)
  *	[0-9]+-[0-9]+
  *
  * treat it as an ascending range of file descriptor numbers.
+ *
+ * Accept a leading '^' as an excusion on match.
  */
 	for (cp1 = fc, err = 0; *cp1;) {
+	    if (*cp1 == '^') {
+		excl = 1;
+		cp1++;
+	    } else
+		excl = 0;
 	    for (cp2 = cp1, dash = (char *)NULL; *cp2 && *cp2 != ','; cp2++) {
 		if (*cp2 == '-')
 		    dash = cp2;
@@ -732,10 +771,14 @@ enter_fd(f)
 		if (dash) {
 		    if (ckfd_range(cp1, dash, cp2, &lo, &hi))
 			err = 1;
-		    else
-			(void) enter_fd_lst((char *)NULL, lo, hi);
-		} else
-		    (void) enter_fd_lst(cp1, 0, 0);
+		    else {
+			if (enter_fd_lst((char *)NULL, lo, hi, excl))
+			    err = 1;
+		    }
+		} else {
+		    if (enter_fd_lst(cp1, 0, 0, excl))
+			err = 1;
+		}
 	    }
 	    if (c == '\0')
 		break;
@@ -750,15 +793,48 @@ enter_fd(f)
  * enter_fd_lst() - make an entry in the FD list, Fdl
  */
 
-static void
-enter_fd_lst(nm, lo, hi)
+static int
+enter_fd_lst(nm, lo, hi, excl)
 	char *nm;			/* FD name (none if NULL) */
 	int lo;				/* FD low boundary (if nm NULL) */
 	int hi;				/* FD high boundary (if nm NULL) */
+	int excl;			/* exclusion on match */
 {
-	char *cp;
-	struct fd_lst *f;
+	char buf[256], *cp;
 	int n;
+	struct fd_lst *f, *ft;
+/*
+ * Don't allow a mixture of exclusions and inclusions.
+ */
+	if (FdlTy >= 0) {
+	    if (FdlTy != excl) {
+		if (!Fwarn) {
+
+		/*
+		 * If warnings are enabled, report a mixture.
+		 */
+		    if (nm) {
+			(void) snpf(buf, sizeof(buf) - 1, "%s%s",
+			    excl ? "^" : "", nm);
+		    } else {
+			if (lo != hi) {
+			    (void) snpf(buf, sizeof(buf) - 1, "%s%d-%d",
+				excl ? "^" : "", lo, hi);
+			} else {
+			    (void) snpf(buf, sizeof(buf) - 1, "%s%d",
+				excl ? "^" : "", lo);
+			}
+		    }
+		    buf[sizeof(buf) - 1] = '\0';
+		    (void) fprintf(stderr,
+		        "%s: %s in an %s -d list: %s\n", Pn,
+			excl ? "exclude" : "include",
+			FdlTy ? "exclude" : "include",
+			buf);
+		}
+		return(1);
+	    }
+	}
 /*
  * Allocate an fd_lst entry.
  */
@@ -795,10 +871,27 @@ enter_fd_lst(nm, lo, hi)
 	    }
 	} else
 	    f->nm = (char *)NULL;
+/*
+ * Skip duplicates.
+ */
+	for (ft = Fdl; ft; ft = ft->next) {
+	    if (f->nm) {
+		if (!ft->nm || strcmp(f->nm, ft->nm))
+		    continue;
+	    } else if ((lo != ft->lo) || (hi != ft->hi))
+		continue;
+	    (void) free((FREE_P *)f);
+	    return(0);
+	}
+/*
+ * Complete the fd_lst entry and link it to the head of the chain.
+ */
 	f->hi = hi;
 	f->lo = lo;
 	f->next = Fdl;
 	Fdl = f;
+	FdlTy = excl;
+	return(0);
 }
 
 
@@ -875,7 +968,7 @@ enter_dir(d, descend)
 	Dstkn = Dstkx = 0;
 	Dstk = (char **)NULL;
 	(void) stkdir(dn);
-	av[0] = (dn == d) ? mkstrcpy(dn, (MALLOC_S)NULL) : dn;
+	av[0] = (dn == d) ? mkstrcpy(dn, (MALLOC_S *)NULL) : dn;
 	av[1] = (char *)NULL;
 	dn = (char *)NULL;
 	if (!ck_file_arg(0, 1, av, 1, 1, &sb)) {
@@ -975,8 +1068,6 @@ enter_dir(d, descend)
 	    /*
 	     * Lstatsafely() the entry; complain if that fails.
 	     *
-	     * Ignore symbolic links and files not not the directory's device.
-	     *
 	     * Stack entries that represent subdirectories.
 	     */
 		if (lstatsafely(fp, &sb)) {
@@ -995,8 +1086,39 @@ enter_dir(d, descend)
 		(void) HASSPECDEVD(fp, &sb);
 #endif	/* defined(HASSPECDEVD) */
 
-		if ((sb.st_mode & S_IFMT) == S_IFLNK || sb.st_dev != ddev)
-		    continue;
+		if (!(Fxover & XO_FILESYS)) {
+
+		/*
+		 * Unless "-x" or "-x f" was specified, don't cross over file
+		 * system mount points.
+		 */
+		    if (sb.st_dev != ddev)
+			continue;
+		}
+		if ((sb.st_mode & S_IFMT) == S_IFLNK) {
+
+		/*
+		 * If this is a symbolic link and "-x_ or "-x l" was specified,
+		 * Statsafely() the entry and process it.
+		 *
+		 * Otherwise skip symbolic links.
+		 */
+		    if (Fxover & XO_SYMLINK) {
+			if (statsafely(fp, &sb)) {
+			    if ((en = errno) != ENOENT) {
+				if (!Fwarn) {
+				    (void) fprintf(stderr,
+					"%s: WARNING: can't stat(", Pn);
+				    safestrprt(fp, stderr, 0);
+				    (void) fprintf(stderr,
+					") symbolc link: %s\n", strerror(en));
+				}
+			    }
+			    continue;
+		        }
+		    } else
+			continue;
+		}
 		if (av[0]) {
 		    (void) free((FREE_P *)av[0]);
 		    av[0] = (char *)NULL;
@@ -1068,7 +1190,7 @@ enter_id(ty, p)
 	char *p;			/* process group ID string pointer */
 {
 	char *cp;
-	int i, id, mx, n;
+	int err, i, id, j, mx, n, ni, nx, x;
 	struct int_lst *s;
 
 	if (!p) {
@@ -1083,11 +1205,15 @@ enter_id(ty, p)
 	case PGID:
 	    mx = Mxpgid;
 	    n = Npgid;
+	    ni = Npgidi;
+	    nx = Npgidx;
 	    s = Spgid;
 	    break;
 	case PID:
 	    mx = Mxpid;
 	    n = Npid;
+	    ni = Npidi;
+	    nx = Npidx;
 	    s = Spid;
 	    break;
 	default:
@@ -1099,12 +1225,19 @@ enter_id(ty, p)
 /*
  * Convert and store the ID.
  */
-	for (cp = p; *cp;) {
+	for (cp = p, err = 0; *cp;) {
 
 	/*
 	 * Assemble ID.
 	 */
-	    for (id = 0; *cp && *cp != ','; cp++) {
+	    for (i = id = x = 0; *cp && *cp != ','; cp++) {
+		if (!i) {
+		    i = 1;
+		    if (*cp == '^') {
+			x = 1;
+			continue;
+		    }
+		}
 
 #if	defined(__STDC__)
 		if (!isdigit((unsigned char)*cp))
@@ -1123,13 +1256,24 @@ enter_id(ty, p)
 	    if (*cp)
 		cp++;
 	/*
-	 * Avoid entering duplicates.
+	 * Avoid entering duplicates and conflicts.
 	 */
-	    for (i = 0; i < n; i++) {
-		if (id == s[i].i)
+	    for (i = j = 0; i < n; i++) {
+		if (id == s[i].i) {
+		    if (x == s[i].x) {
+			j = 1;
+			continue;
+		    }
+		    (void) fprintf(stderr,
+			"%s: P%sID %d has been included and excluded.\n",
+			Pn,
+			(ty == PGID) ? "G" : "",
+			id);
+		    err = j = 1;
 		    break;
+		}
 	    }
-	    if (i < n)
+	    if (j)
 		continue;
 	/*
 	 * Allocate table table space.
@@ -1149,7 +1293,12 @@ enter_id(ty, p)
 		}
 	    }
 	    s[n].f = 0;
-	    s[n++].i = id;
+	    s[n].i = id;
+	    s[n++].x = x;
+	    if (x)
+		nx++;
+	    else
+		ni++;
 	}
 /*
  * Save variables for the type of ID.
@@ -1157,13 +1306,17 @@ enter_id(ty, p)
 	if (ty == PGID) {
 	    Mxpgid = mx;
 	    Npgid = n;
+	    Npgidi = ni;
+	    Npgidx = nx;
 	    Spgid = s;
 	} else {
 	    Mxpid = mx;
 	    Npid = Npuns = n;
+	    Npidi = ni;
+	    Npidx = nx;
 	    Spid = s;
 	}
-	return(0);
+	return(err);
 }
 
 
@@ -1177,6 +1330,7 @@ enter_network_address(na)
 {
 	int ae, i, pr;
 	int ep = -1;
+	int ft = 0;
 	struct hostent *he = (struct hostent *)NULL;
 	char *hn = (char *)NULL;
 	MALLOC_S l;
@@ -1197,12 +1351,69 @@ enter_network_address(na)
 	    (void) fprintf(stderr, "%s: no network address specified\n", Pn);
 	    return(1);
 	}
+	zeromem((char *)&n, sizeof(n));
+	wa = na;
+/*
+ * Process an IP version type specification, IPv4 or IPv6, optionally followed
+ * by a '@' and a host name or Internet address, or a ':' and a service name or
+ * port number.
+ */
+	if ((*wa == '4') || (*wa == '6')) {
+	    if (*wa == '4')
+		ft = 4;
+	    else if (*wa == '6') {
+
+#if	defined(HASIPv6)
+		ft = 6;
+#else	/* !defined(HASIPv6) */
+		(void) fprintf(stderr, "%s: IPv6 not supported: -i ", Pn);
+		safestrprt(na, stderr, 1);
+		goto nwad_exit;
+#endif	/* defined(HASIPv6) */
+
+	    }
+	    wa++;
+	    if (!*wa) {
+
+	    /*
+	     * If nothing follows 4 or 6, then all network files of the
+	     * specified IP version are selected.  Sequential -i, -i4, and
+	     * -i6 specifications interact logically -- e.g., -i[46] followed
+	     * by -i[64] is the same as -i.
+	     */
+		if (!Fnet) {
+		    Fnet = 1;
+		    FnetTy = ft;
+		} else {
+		    if (FnetTy) {
+			if (FnetTy != ft)
+			    FnetTy = 0;
+		    } else
+			FnetTy = ft;
+		}
+		return(0);
+	    }
+	} else if (Fnet)
+	    ft = FnetTy;
+/*
+ * If an IP version has been specified, use it to set the address family.
+ */
+	switch (ft) {
+	case 4:
+	    n.af = AF_INET;
+	    break;
+
+#if	defined(HASIPv6)
+	case 6:
+	    n.af = AF_INET6;
+	    break;
+#endif	/* defined(HASIPv6) */
+
+	}
 /*
  * Process protocol name, optionally followed by a '@' and a host name or
  * Internet address, or a ':' and a service name or port number.
  */
-	zeromem((char *)&n, sizeof(n));
-	wa = na;
 	if (*wa && *wa != '@' && *wa != ':') {
 	    for (p = wa; *wa && *wa != '@' && *wa != ':'; wa++)
 		;
@@ -1211,7 +1422,7 @@ enter_network_address(na)
 			        -1, (MALLOC_S *)NULL)))
 		{
 		    (void) fprintf(stderr,
-			"%s: no space for protocol name from ", Pn);
+			"%s: no space for protocol name from: -i ", Pn);
 		    safestrprt(na, stderr, 1);
 nwad_exit:
 		    if (n.proto)
@@ -1223,13 +1434,14 @@ nwad_exit:
 		    return(1);
 		}
 	    /*
-	     * The protocol name should be "tcp" or "udp".
+	     * The protocol name should be "tcp", "udp" or "udplite".
 	     */
-		if (strcasecmp(n.proto, "tcp") != 0
-		&&  strcasecmp(n.proto, "udp") != 0)
+		if ((strcasecmp(n.proto, "tcp") != 0)
+		&&  (strcasecmp(n.proto, "udp") != 0)
+		&&  (strcasecmp(n.proto, "udplite") != 0))
 		{
 		    (void) fprintf(stderr,
-			"%s: unknown protocol name (%s) in ", Pn, n.proto);
+			"%s: unknown protocol name (%s) in: -i ", Pn, n.proto);
 		    safestrprt(na, stderr, 1);
 		    goto nwad_exit;
 		}
@@ -1251,58 +1463,42 @@ nwad_exit:
 	    wa++;
 	    if (!*wa || *wa == ':') {
 
+#if	defined(HASIPv6)
 unacc_address:
+#endif	/* defined(HASIPv6) */
+
 		(void) fprintf(stderr,
-		    "%s: unacceptable Internet address in ", Pn);
+		    "%s: unacceptable Internet address in: -i ", Pn);
 		safestrprt(na, stderr, 1);
 		goto nwad_exit;
 	    }
-	    if ((*wa < '0' || *wa > '9') && *wa != '[') {
+
+	    if ((p = isIPv4addr(wa, n.a, sizeof(n.a)))) {
 
 	    /*
-	     * Assemble host name.
+	     * Process IPv4 address.
 	     */
-		for (p = wa; *p && *p != ':'; p++)
-		    ;
-		if ((l = p - wa)) {
-		    if (!(hn = mkstrcat(wa, l, (char *)NULL, -1, (char *)NULL,
-			       -1, (MALLOC_S *)NULL)))
-		    {
-			(void) fprintf(stderr,
-			    "%s: no space for host name: ", Pn);
-			safestrprt(na, stderr, 1);
-			goto nwad_exit;
-		    }
-
-#if	defined(HASIPv6)
-		    n.af = AF_INET6;
-#else	/* !defined(HASIPv6) */
-		    n.af = AF_INET;
-#endif	/* defined(HASIPv6) */
-
-		    he = lkup_hostnm(hn, &n);
-
-#if	defined(HASIPv6)
-		    if (!he) {
-			n.af = AF_INET;
-			he = lkup_hostnm(hn, &n);
-		    }
-#endif	/* defined(HASIPv6) */
-		
-		    if (!he) {
-			fprintf(stderr, "%s: unknown host name (%s) in ",
-			    Pn, hn);
-			safestrprt(na, stderr, 1);
-			goto nwad_exit;
-		    }
+		if (ft == 6) {
+		    (void) fprintf(stderr,
+			"%s: IPv4 addresses are prohibited: -i ", Pn);
+		    safestrprt(na, stderr, 1);
+		    goto nwad_exit;
 		}
 		wa = p;
+		n.af = AF_INET;
 	    } else if (*wa == '[') {
 
 #if	defined(HASIPv6)
 	    /*
-	     * Assemble IPv6 address.
+	     * Make sure IPv6 addresses are permitted.  If they are, assemble
+	     * one.
 	     */
+		if (ft == 4) {
+		    (void) fprintf(stderr,
+			"%s: IPv6 addresses are prohibited: -i ", Pn);
+		    safestrprt(na, stderr, 1);
+		    goto nwad_exit;
+		}
 		if (!(cp = strrchr(++wa, ']')))
 		    goto unacc_address;
 		*cp = '\0';
@@ -1317,6 +1513,12 @@ unacc_address:
 		if (!ae)
 		    goto unacc_address;
 		if (IN6_IS_ADDR_V4MAPPED((struct in6_addr *)&n.a[0])) {
+		    if (ft == 6) {
+			(void) fprintf(stderr,
+			    "%s: IPv4 addresses are prohibited: -i ", Pn);
+			safestrprt(na, stderr, 1);
+			goto nwad_exit;
+		    }
 		    for (i = 0; i < 4; i++) {
 			n.a[i] = n.a[i+12];
 		    }
@@ -1326,37 +1528,57 @@ unacc_address:
 		wa = cp + 1;
 #else	/* !defined(HASIPv6) */
 		(void) fprintf(stderr,
-		    "%s: unsupported IPv6 address in ", Pn);
+		    "%s: unsupported IPv6 address in: -i ", Pn);
 		safestrprt(na, stderr, 1);
 		goto nwad_exit;
 #endif	/* defined(HASIPv6) */
 
-
 	    } else {
 
 	    /*
-	     * Assemble IPv4 address.
+	     * Assemble host name.
 	     */
-		for (i = 0; *wa; wa++) {
-		    if (*wa == ':')
-			break;
-		    if (*wa == '.') {
-			i++;
-			if (i >= MIN_AF_ADDR)
-			    break;
-			continue;
+		for (p = wa; *p && *p != ':'; p++)
+		    ;
+		if ((l = p - wa)) {
+		    if (!(hn = mkstrcat(wa, l, (char *)NULL, -1, (char *)NULL,
+			       -1, (MALLOC_S *)NULL)))
+		    {
+			(void) fprintf(stderr,
+			    "%s: no space for host name: -i ", Pn);
+			safestrprt(na, stderr, 1);
+			goto nwad_exit;
+		    }
+
+#if	defined(HASIPv6)
+
+		/*
+		 * If no IP version has been specified, look up an IPv6 host
+		 * name first.  If that fails, look up an IPv4 host name.
+		 *
+		 * If the IPv6 version has been specified, look up the host
+		 * name only under its IP version specification.
+		 */
+		    if (!ft)
+			n.af = AF_INET6;
+		    if (!(he = lkup_hostnm(hn, &n)) && !ft) {
+			n.af = AF_INET;
+			he = lkup_hostnm(hn, &n);
+		    }
+#else	/* !defined(HASIPv6) */
+		    if (!ft)
+			n.af = AF_INET;
+		    he = lkup_hostnm(hn, &n);
+#endif	/* defined(HASIPv6) */
+		
+		    if (!he) {
+			fprintf(stderr, "%s: unknown host name (%s) in: -i ",
+			    Pn, hn);
+			safestrprt(na, stderr, 1);
+			goto nwad_exit;
 		    }
-		    if (*wa < '0' || *wa > '9')
-			goto unacc_address;
-		    ae = (10 * n.a[i]) + *wa - '0';
-		    if (ae > 255)
-			goto unacc_address;
-		    n.a[i] = (unsigned char)ae;
 		}
-		if (i != (MIN_AF_ADDR - 1)
-		||  (!n.a[0] && !n.a[1] && !n.a[2] && !n.a[3]))
-		    goto unacc_address;
-		n.af = AF_INET;
+		wa = p;
 	    }
 	}
 /*
@@ -1377,7 +1599,7 @@ unacc_address:
 
 unacc_port:
 	    (void) fprintf(stderr,
-		"%s: unacceptable port specification in ", Pn);
+		"%s: unacceptable port specification in: -i ", Pn);
 	    safestrprt(na, stderr, 1);
 	    goto nwad_exit;
 	}
@@ -1388,13 +1610,13 @@ unacc_port:
 		/*
 		 * Convert service name to port number, using already-specified
 		 * protocol name.  A '-' is taken to be part of the name; hence
-		 * the staring entry of a range can't be a service name.
+		 * the starting entry of a range can't be a service name.
 		 */
 		    for (p = wa; *wa && *wa != ','; wa++)
 			;
 		    if (!(l = wa - p)) {
 			(void) fprintf(stderr,
-			    "%s: invalid service name: ", Pn);
+			    "%s: invalid service name: -i ", Pn);
 			safestrprt(na, stderr, 1);
 			goto nwad_exit;
 		    }
@@ -1409,7 +1631,7 @@ unacc_port:
 		    }
 		    if (!sn) {
 			(void) fprintf(stderr,
-			    "%s: no space for service name: ", Pn);
+			    "%s: no space for service name: -i ", Pn);
 			safestrprt(na, stderr, 1);
 			goto nwad_exit;
 		    }
@@ -1423,7 +1645,7 @@ unacc_port:
 		     */
 			if (!(se = getservbyname(sn, n.proto))) {
 			    (void) fprintf(stderr,
-				"%s: unknown service %s for %s in ",
+				"%s: unknown service %s for %s in: -i ",
 				Pn, sn, n.proto);
 			    safestrprt(na, stderr, 1);
 			    goto nwad_exit;
@@ -1441,7 +1663,7 @@ unacc_port:
 			    pu = (int)ntohs(se1->s_port);
 			if (!se && !se1) {
 			    (void) fprintf(stderr,
-				"%s: unknown service %s in ", Pn, sn);
+				"%s: unknown service %s in: -i ", Pn, sn);
 			    safestrprt(na, stderr, 1);
 			    goto nwad_exit;
 			}
@@ -1450,8 +1672,9 @@ unacc_port:
 				"%s: TCP=%d and UDP=%d %s ports conflict;\n",
 				Pn, pt, pu, sn);
 			    (void) fprintf(stderr,
-				"      specify \"tcp:%s\" or \"udp:%s\".\n",
+				"      specify \"tcp:%s\" or \"udp:%s\": -i ",
 				sn, sn);
+			    safestrprt(na, stderr, 1);
 			    goto nwad_exit;
 			}
 			if (!se && se1)
@@ -1508,9 +1731,10 @@ nwad_enter:
 	    /*
 	     * If IPv6 is enabled, a host name was specified, and the
 	     * associated * address is for the AF_INET6 address family,
-	     * try to get and address for the AF_INET family, too.
+	     * try to get and address for the AF_INET family, too, unless
+	     * IPv4 is prohibited.
 	     */
-		if (hn && (n.af == AF_INET6)) {
+		if (hn && (n.af == AF_INET6) && (ft != 6)) {
 		    n.af = AF_INET;
 		    if ((he = lkup_hostnm(hn, &n)))
 			continue;
@@ -1552,7 +1776,7 @@ enter_nwad(n, sp, ep, s, he)
 	if (strlen(s)) {
 	    if (!(n->arg = mkstrcpy(s, (MALLOC_S *)NULL))) {
 		(void) fprintf(stderr,
-		    "%s: no space for Internet argument: ", Pn);
+		    "%s: no space for Internet argument: -i ", Pn);
 		safestrprt(s, stderr, 1);
 		Exit(1);
 	    }
@@ -1580,7 +1804,7 @@ enter_nwad(n, sp, ep, s, he)
 
 	    &&  sp == -1) {
 		(void) fprintf(stderr,
-		    "%s: incomplete Internet address specification: ", Pn);
+		    "%s: incomplete Internet address specification: -i ", Pn);
 		safestrprt(s, stderr, 1);
 		return(1);
 	    }
@@ -1590,7 +1814,7 @@ enter_nwad(n, sp, ep, s, he)
 	 */
 	    if (na >= MAXNWAD) {
 		(void) fprintf(stderr,
-		    "%s: network address limit (%d) exceeded: ",
+		    "%s: network address limit (%d) exceeded: -i ",
 		    Pn, MAXNWAD);
 		safestrprt(s, stderr, 1);
 		return(1);
@@ -1600,7 +1824,7 @@ enter_nwad(n, sp, ep, s, he)
 	 */
 	    if ((np = (struct nwad *)malloc(sizeof(struct nwad))) == NULL) {
 		(void) fprintf(stderr,
-		    "%s: no space for network address from: ", Pn);
+		    "%s: no space for network address from: -i ", Pn);
 		safestrprt(s, stderr, 1);
 		return(1);
 	    }
@@ -1647,17 +1871,242 @@ enter_nwad(n, sp, ep, s, he)
 }
 
 
+#if	defined(HASTCPUDPSTATE)
+/*
+ * enter_state_spec() -- enter TCP and UDP state specifications
+ */
+
+int
+enter_state_spec(ss)
+	char *ss;			/* state specification string */
+{
+	char *cp, *ne, *ns, *pr;
+	int err, d, f, i, tx, x;
+	size_t len;
+	static char *ssc = (char *)NULL;
+	char *ty;
+/*
+ * Check the protocol specification.
+ */
+	if (!strncasecmp(ss, "tcp:", 4)) {
+	    pr = "TCP";
+	    tx = 0;
+	}
+
+#if	!defined(USE_LIB_PRINT_TCPTPI)
+	else if (!strncasecmp(ss, "UDP:", 4)) {
+	    pr = "UDP";
+	    tx = 1;
+	}
+
+#endif	/* !defined(USE_LIB_PRINT_TCPTPI) */
+
+	else {
+	    (void) fprintf(stderr, "%s: unknown -s protocol: \"%s\"\n",
+		Pn, ss);
+	    return(1);
+	}
+	cp = ss + 4;
+	if (!*cp) {
+	    (void) fprintf(stderr, "%s: no %s state names in: %s\n",
+		Pn, pr, ss);
+	    return(1);
+	}
+	(void) build_IPstates();
+	if (!(tx ? UdpSt : TcpSt)) {
+	    (void) fprintf(stderr, "%s: no %s state names available: %s\n",
+		Pn, pr, ss);
+	    return(1);
+	}
+/*
+ * Allocate the inclusion and exclusion tables for the protocol.
+ */
+	if (tx) {
+	    if (UdpNstates) {
+		if (!UdpStI) {
+		    if (!(UdpStI = (unsigned char *)calloc((MALLOC_S)UdpNstates,
+				   sizeof(unsigned char))))
+		    {
+			ty = "UDP state inclusion";
+
+no_IorX_space:
+
+			(void) fprintf(stderr, "%s: no %s table space\n",
+			    Pn, ty);
+			Exit(1);
+		    }
+		}
+		if (!UdpStX) {
+		    if (!(UdpStX = (unsigned char *)calloc((MALLOC_S)UdpNstates,
+				   sizeof(unsigned char))))
+		    {
+			ty = "UDP state exclusion";
+			goto no_IorX_space;
+		    }
+		}
+	    }
+	} else {
+	    if (TcpNstates) {
+		if (!TcpStI) {
+		    if (!(TcpStI = (unsigned char *)calloc((MALLOC_S)TcpNstates,
+				   sizeof(unsigned char))))
+		    {
+			ty = "TCP state inclusion";
+			goto no_IorX_space;
+		    }
+		}
+		if (!TcpStX) {
+		    if (!(TcpStX = (unsigned char *)calloc((MALLOC_S)TcpNstates,
+				   sizeof(unsigned char))))
+		    {
+			ty = "TCP state exclusion";
+			goto no_IorX_space;
+		    }
+		}
+	    }
+	}
+/*
+ * Convert the state names in the rest of the string to state indexes and
+ * record them in the appropriate inclusion or exclusion table.
+ */
+	if (ssc)
+	    (void) free((MALLOC_P *)ssc);
+	if (!(ssc = mkstrcpy(cp, (MALLOC_S)NULL))) {
+	    (void) fprintf(stderr,
+		"%s: no temporary state argument space for: %s\n", Pn, ss);
+	    Exit(1);
+	}
+	cp = ssc;
+	err = 0;
+	while (*cp) {
+	
+	/*
+	 * Determine inclusion or exclusion for this state name.
+	 */
+	    if (*cp == '^') {
+		x = 1;
+		cp++;
+	    } else
+		x = 0;
+	/*
+	 * Find the end of the state name.  Make sure it is non-null in length
+	 * and terminated with '\0'.
+	 */
+	    ns = cp;
+	    while (*cp && (*cp != ',')) {
+		cp++;
+	    }
+	    ne = cp;
+	    if (*cp) {
+		*cp = '\0';
+		cp++;
+	    }
+	    if (!(len = (size_t)(ne - ns))) {
+		(void) fprintf(stderr, "%s: NULL %s state name in: %s\n",
+		    Pn, pr, ss);
+		err = 1;
+		continue;
+	    }
+	/*
+	 * Find the state name in the appropriate table.
+	 */
+	    f = 0;
+	    if (tx) {
+		if (UdpSt) {
+		    for (i = 0; i < UdpNstates; i++) {
+			if (!strcasecmp(ns, UdpSt[i])) {
+			    f = 1;
+			    break;
+			}
+		    }
+		}
+	    } else {
+		if (TcpSt) {
+		    for (i = 0; i < TcpNstates; i++) {
+			if (!strcasecmp(ns, TcpSt[i])) {
+			    f = 1;
+			    break;
+			}
+		    }
+		}
+	    }
+	    if (!f) {
+		(void) fprintf(stderr, "%s: unknown %s state name: %s\n",
+		    Pn, pr, ns);
+		err = 1;
+		continue;
+	    }
+	/*
+	 * Set the inclusion or exclusion status in the appropriate table.
+	 */
+	    d = 0;
+	    if (x) {
+		if (tx) {
+		    if (!UdpStX[i]) {
+			UdpStX[i] = 1;
+			UdpStXn++;
+		    } else
+			d = 1;
+		} else {
+		    if (!TcpStX[i]) {
+			TcpStX[i] = 1;
+			TcpStXn++;
+		    } else
+			d = 1;
+		}
+	    } else {
+		if (tx) {
+		    if (!UdpStI[i]) {
+			UdpStI[i] = 1;
+			UdpStIn++;
+		    } else
+			d = 1;
+		} else {
+		    if (!TcpStI[i]) {
+			TcpStI[i] = 1;
+			TcpStIn++;
+		    } else
+			d = 1;
+		}
+	    }
+	    if (d) {
+
+	    /*
+	     * Report a duplicate.
+	     */
+		(void) fprintf(stderr, "%s: duplicate %s %sclusion: %s\n",
+		    Pn, pr,
+		    x ? "ex" : "in",
+		    ns);
+		err = 1;
+	    }
+	}
+/*
+ * Release any temporary space and return.
+ */
+	if (ssc) {
+	    (void) free((MALLOC_P *)ssc);
+	    ssc = (char *)NULL;
+	}
+	return(err);
+}
+#endif	/* defined(HASTCPUDPSTATE) */
+
+
 /*
  * enter_str_lst() - enter a string on a list
  */
 
 int
-enter_str_lst(opt, s, lp)
+enter_str_lst(opt, s, lp, incl, excl)
 	char *opt;			/* option name */
 	char *s;			/* string to enter */
 	struct str_lst **lp;		/* string's list */
+	int *incl;			/* included count */
+	int *excl;			/* excluded count */
 {
 	char *cp;
+	short i, x;
 	MALLOC_S len;
 	struct str_lst *lpt;
 
@@ -1666,6 +2115,14 @@ enter_str_lst(opt, s, lp)
 		Pn, opt);
 	    return(1);
 	}
+	if (*s == '^') {
+	    i = 0;
+	    x = 1;
+	    s++;
+	} else {
+	    i = 1;
+	    x = 0;
+	}
 	if (!(cp = mkstrcpy(s, &len))) {
 	    (void) fprintf(stderr, "%s: no string copy space: ", Pn);
 	    safestrprt(s, stderr, 1);
@@ -1680,6 +2137,11 @@ enter_str_lst(opt, s, lp)
 	lpt->f = 0;
 	lpt->str = cp;
 	lpt->len = (int)len;
+	lpt->x = x;
+	if (i)
+	    *incl += 1;
+	if (x)
+	    *excl += 1;
 	lpt->next = *lp;
 	*lp = lpt;
 	return(0);
@@ -1721,7 +2183,8 @@ enter_uid(us)
 			lnml++;
 		    }
 		    (void) fprintf(stderr,
-			"%s: -u login name > %d characters: ", Pn, LOGINML);
+			"%s: -u login name > %d characters: ", Pn,
+			    (int)LOGINML);
 		    safestrprtn(st, lnml, stderr, 1);
 		    err = j = 1;
 		    break;
@@ -1759,11 +2222,12 @@ enter_uid(us)
 		    uid = pw->pw_uid;
 	    }
 
-#if	defined(HASSECURITY)
+#if	defined(HASSECURITY) && !defined(HASNOSOCKSECURITY)
 	/*
 	 * If the security mode is enabled, only the root user may list files
 	 * belonging to user IDs other than the real user ID of this lsof
-	 * process.
+	 * process.  If HASNOSOCKSECURITY is also defined, then anyone may
+	 * list anyone else's socket files.
 	 */
 	    if (Myuid && uid != Myuid) {
 		(void) fprintf(stderr,
@@ -1772,7 +2236,7 @@ enter_uid(us)
 		err = 1;
 		continue;
 	    }
-#endif	/* HASSECURITY */
+#endif	/* defined(HASSECURITY)  && !defined(HASNOSOCKSECURITY) */
 
 	/*
 	 * Avoid entering duplicates.
@@ -1785,7 +2249,7 @@ enter_uid(us)
 		    continue;
 		}
 		(void) fprintf(stderr,
-		    "%s: UID %d has been included and excluded\n",
+		    "%s: UID %d has been included and excluded.\n",
 			Pn, (int)uid);
 		err = j = 1;
 		break;
@@ -1808,7 +2272,7 @@ enter_uid(us)
 		}
 	    }
 	    if (nn) {
-		if (!(lp = mkstrcpy(lnm, (MALLOC_S)NULL))) {
+		if (!(lp = mkstrcpy(lnm, (MALLOC_S *)NULL))) {
 		    (void) fprintf(stderr, "%s: no space for login: ", Pn);
 		    safestrprt(lnm, stderr, 1);
 		    Exit(1);
@@ -1827,6 +2291,82 @@ enter_uid(us)
 }
 
 
+/*
+ * isIPv4addr() - is host name an IPv4 address
+ */
+
+static char *
+isIPv4addr(hn, a, al)
+	char *hn;			/* host name */
+	unsigned char *a;		/* address receptor */
+	int al;				/* address receptor length */
+{
+	int dc = 0;			/* dot count */
+	int i;				/* temorary index */
+	int ov[MIN_AF_ADDR];		/* octet values */
+	int ovx = 0;			/* ov[] index */
+/*
+ * The host name must begin with a number and the return octet value
+ * arguments must be acceptable.
+ */
+	if ((*hn < '0') || (*hn > '9'))
+	    return((char *)NULL);
+	if (!a || (al < MIN_AF_ADDR))
+	    return((char *)NULL);
+/*
+ * Start the first octet assembly, then parse tge remainder of the host
+ * name for four octets, separated by dots.
+ */
+	ov[0] = (int)(*hn++ - '0');
+	while (*hn && (*hn != ':')) {
+	    if (*hn == '.') {
+
+	    /*
+	     * Count a dot.  Make sure a preceding octet value has been
+	     * assembled.  Don't assemble more than MIN_AF_ADDR octets.
+	     */
+		dc++;
+		if ((ov[ovx] < 0) || (ov[ovx] > 255))
+		    return((char *)NULL);
+		if (++ovx > (MIN_AF_ADDR - 1))
+		    return((char *)NULL);
+		ov[ovx] = -1;
+	    } else if ((*hn >= '0') && (*hn <= '9')) {
+
+	    /*
+	     * Assemble an octet.
+	     */
+		if (ov[ovx] < 0)
+		    ov[ovx] = (int)(*hn - '0');
+		else
+		    ov[ovx] = (ov[ovx] * 10) + (int)(*hn - '0');
+	    } else {
+
+	    /*
+	     * A non-address character has been detected.
+	     */
+		return((char *)NULL);
+	    }
+	    hn++;
+	}
+/*
+ * Make sure there were three dots and four non-null octets.
+ */
+	if ((dc != 3)
+	||  (ovx != (MIN_AF_ADDR - 1))
+	||  (ov[ovx] < 0) || (ov[ovx] > 255))
+	    return((char *)NULL);
+/*
+ * Copy the octets as unsigned characters and return the ending host name
+ * character position.
+ */
+	for (i = 0; i < MIN_AF_ADDR; i++) {
+	     a[i] = (unsigned char)ov[i];
+	}
+	return(hn);
+}
+
+
 /*
  * lkup_hostnm() - look up host name
  */
@@ -1885,5 +2425,4 @@ lkup_hostnm(hn, n)
 	if ((ln = MAX_AF_ADDR - 4) > 0)
 	    zeromem((char *)&n->a[4], ln);
 	return(he);
-
 }
