@@ -31,7 +31,7 @@
 
 
 /*
- * $Id: dlsof.h,v 1.36 2004/12/30 18:41:31 abe Exp $
+ * $Id: dlsof.h,v 1.41 2008/10/21 16:16:06 abe Exp $
  */
 
 
@@ -43,15 +43,15 @@
 #include <nlist.h>
 #include <setjmp.h>
 #include <signal.h>
-#include <string.h>
 #include <unistd.h>
 
 # if	FREEBSDV>=4000
 #  if	FREEBSDV>=5000
+#   if	FREEBSDV<6020
 #    if	defined(__alpha__)
 /*
- * For Alpha, #include <machine/pcpu.h> before #define'ing _KERNEL.  Then
- * #define PCPU_MD_FIELDS independently.  This hack avoids a compiler
+ * For Alpha below 6.2, #include <machine/pcpu.h> before #define'ing _KERNEL.
+ * Then #define PCPU_MD_FIELDS independently.  This hack avoids a compiler
  * complaint about register use.
  */
 
@@ -63,9 +63,16 @@
 	u_int32_t       pc_next_asn;            /* next ASN to alloc */ \
 	u_int32_t       pc_current_asngen       /* ASN rollover check */
 #    endif	/* defined(__alpha__) */
+#   endif	/* FREEBSDV<6020 */
 #define	_KERNEL	1
 #  endif	/* FREEBSDV>=5000 */
 #include <sys/conf.h>
+
+#  if	defined(HAS_CONF_MINOR)
+#undef	minor
+#include "fbsd_minor.h"
+#  endif	/* defined(HAS_CONF_MINOR) */
+
 #  if	FREEBSDV>=5000
 #undef	_KERNEL
 #  endif	/* FREEBSDV>=5000 */
@@ -143,6 +150,20 @@
 
 #include <ufs/ufs/inode.h>
 
+# if	defined(HAS_UFS1_2)
+#define	_KERNEL
+struct vop_getextattr_args;
+struct vop_deleteextattr_args;
+struct vop_setextattr_args;
+#include <ufs/ufs/extattr.h>
+#define	psignal	LSOF_psignal
+#define	panicstr bp
+#include <ufs/ufs/ufsmount.h>
+#undef	psignal
+#undef	panicstr
+#undef	_KERNEL
+# endif	/* defined(HAS_UFS1_2) */
+
 #  if	FREEBSDV>=5010
 #undef	i_devvp
 #  endif	/* FREEBSDV>=5010 */
@@ -230,6 +251,10 @@ struct vop_generic_args;
 #include <fs/pseudofs/pseudofs.h>
 # endif	/* defined(HASPSEUDOFS) */
 
+# if	defined(HAS_ZFS)
+#include "dzfs.h"
+# endif	/* defined(HAS_ZFS) */
+
 
 # if	FREEBSDV<2000
 #define	P_COMM		p_comm
@@ -263,7 +288,35 @@ struct vop_generic_args;
 #define	_KERNEL
 #define	KERNEL
 #include <sys/fcntl.h>
+
+/*
+ * The following circumventions were first needed in FreeBSD 8.0-CURRENT some
+ * time in August 2008 to avoid conflicts in /usr/src/sys/sys/libkern.h> and
+ * /usr/src/sys/sys/systm.h, called by <sys/file.h> or the header files it
+ * #include's when KERNEL or _KERNEL is #define'd.
+ *
+ * The circumventions may be needed or may be erroneous for earlier FreeBSD
+ * versions where testing was not possible.
+ */
+
+#define	intrmask_t	int
+#define	log	log_kernel_lsof
+#define	pause	pause_kernel_lsof
+#define	setenv	setenv_kernel_lsof
+#define	uintfptr_t	int
+#define	_SYS_LIBKERN_H_
 #include <sys/file.h>
+
+/*
+ * Attempt to remove the circumventions.
+ */
+
+#undef	_SYS_LIBKERN_H_
+#undef	intrmask_t_lsof
+#undef	log_kernel_lsof
+#undef	pause_kernel_lsof
+#undef	setenv_kernel_lsof
+#undef	uintfptr_t
 #undef	_KERNEL
 #undef	KERNEL
 
@@ -287,6 +340,15 @@ struct vop_advlock_args { int dummy; };	/* to pacify lf_advlock() prototype */
 #define	MALLOC_DECLARE(type)	extern struct malloc_type type[1]
 					/* to pacify <sys/lockf.h> */
 #define	_KERNEL
+
+#   if	defined(HAS_SYS_SX_H)
+#include <sys/sx.h>
+#   endif	/* defined(HAS_SYS_SX_H) */
+
+#   if	defined(HAS_SI_PRIV) || defined(HAS_CONF_MINOR)
+#include <fs/devfs/devfs_int.h>
+#   endif	/* defined(SI_PRIV) || defined(HAS_CONF_MINOR) */
+
 #include <fs/devfs/devfs.h>
 #undef	_KERNEL
 #  endif	/* FREEBSDV>=5000 */
@@ -338,20 +400,25 @@ struct vop_advlock_args { int dummy; };	/* to pacify lf_advlock() prototype */
 #   endif	/* FREEBSDV<5000 */
 #  endif        /* FREEBSDV>=2020 */
 
+#include <string.h>
+
 
 #define	COMP_P		const void
 #define DEVINCR		1024	/* device table malloc() increment */
 
-# if	FREEBSDV<2000
+# if	!defined(FREEBSD_KA_T)
+#  if	FREEBSDV<2000
 typedef	off_t		KA_T;
-# else	/* FREEBSDV>=2000 */
+#  else	/* FREEBSDV>=2000 */
 typedef	u_long		KA_T;
-# endif	/* FREEBSDV<2000 */
+#  endif	/* FREEBSDV<2000 */
+# endif	/* !defined(FREEBSD_KA_T) */
 
 #define	KMEM		"/dev/kmem"
 #define MALLOC_P	void
 #define FREE_P		MALLOC_P
 #define MALLOC_S	size_t
+#define	MAXSYSCMDL	MAXCOMLEN	/* max system command name length */
 
 # if	defined(N_UNIXV)
 #define	N_UNIX_TMP(x)	#x
@@ -360,10 +427,15 @@ typedef	u_long		KA_T;
 # endif	/* defined(N_UNIXV) */
 
 #define QSORT_P		void
+
+# if	!defined(READLEN_T)
 #define	READLEN_T	int
+# endif	/* !defined(READLEN_T) */
+
 #define STRNCPY_L	size_t
 #define SWAP		"/dev/drum"
-#define	SZOFFTYPE	long long	/* size and offset internal storage
+#define	SZOFFTYPE	unsigned long long
+					/* size and offset internal storage
 					 * type */
 #define	SZOFFPSPEC	"ll"		/* SZOFFTYPE print specification
 					 * modifier */
@@ -407,7 +479,7 @@ struct mounts {
 					 * (symbolic links resolved) */
         dev_t dev;              	/* directory st_dev */
 	dev_t rdev;			/* directory st_rdev */
-	ino_t inode;			/* directory st_ino */
+	INODETYPE inode;		/* directory st_ino */
 	mode_t mode;			/* directory st_mode */
 	mode_t fs_mode;			/* file system st_mode */
         struct mounts *next;    	/* forward link */
@@ -432,16 +504,16 @@ struct sfile {
 	u_short mode;			/* S_IFMT mode bits from stat() */
 	int type;			/* file type: 0 = file system
 				 	 *	      1 = regular file */
-	ino_t i;			/* inode number */
+	INODETYPE i;			/* inode number */
 	int f;				/* file found flag */
 	struct sfile *next;		/* forward link */
 
 };
 
-# if	FREEBSDV==4100
+# if	FREEBSDV==4100 || FREEBSDV==4110
 #define	XDR_VOID	(xdrproc_t)xdr_void
 #define	XDR_PMAPLIST	(xdrproc_t)xdr_pmaplist
-# endif	/* FREEBSDV==4100 */
+# endif	/* FREEBSDV==4100 || FREEBSDV==4110 */
 
 # if	FREEBSDV>=5000
 #define	XDR_VOID	(const xdrproc_t)xdr_void
@@ -525,4 +597,4 @@ struct	namecache {
 #define	NCACHE_VROOT	VV_ROOT
 #endif	/* FREEBSDV>=5000 */
 
-#endif	/* FREEBSD_LSOF_H */
+#endif	/* defined(FREEBSD_LSOF_H) */
