@@ -35,7 +35,7 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dnode1.c,v 1.7 2004/07/06 19:18:57 abe Exp $";
+static char *rcsid = "$Id: dnode1.c,v 1.10 2008/10/21 16:16:06 abe Exp $";
 #endif
 
 
@@ -57,6 +57,12 @@ static char *rcsid = "$Id: dnode1.c,v 1.7 2004/07/06 19:18:57 abe Exp $";
 
 #include "cd9660_node.h"
 
+# if	defined(HAS_NO_ISO_DEV)
+#define	_KERNEL
+#include <isofs/cd9660/iso.h>
+#undef	_KERNEL
+# endif	/* defined(HAS_NO_ISO_DEV) */
+
 #  if	FREEBSDV>=4000 && defined(__alpha__)
 #undef	dev_t
 #  endif	/* FREEBSDV>=4000 && defined(__alpha__) */
@@ -71,9 +77,9 @@ read_iso_node(v, d, dd, ino, nl, sz)
 	struct vnode *v;		/* containing vnode */
 	dev_t *d;			/* returned device number */
 	int *dd;			/* returned device-defined flag */
-	unsigned long *ino;		/* returned inode number */
+	INODETYPE *ino;			/* returned inode number */
 	long *nl;			/* returned number of links */
-	unsigned long *sz;		/* returned size */
+	SZOFFTYPE *sz;			/* returned size */
 {
 
 # if	FREEBSDV<2000
@@ -87,6 +93,9 @@ read_iso_node(v, d, dd, ino, nl, sz)
 	struct specinfo udev;
 #  else	/* FREEBSDV>=5000 */
 	struct cdev udev;
+#   if	defined(HAS_NO_ISO_DEV)
+	struct iso_mnt im;
+#   endif	/* defined(HAS_NO_ISO_DEV) */
 #  endif	/* FREEBSDV<5000 */
 # endif	/* FREEBSDV>=4000 */
 
@@ -94,17 +103,30 @@ read_iso_node(v, d, dd, ino, nl, sz)
 	ip = (struct iso_node *)v->v_data;
 	*d = ip->i_dev;
 	*dd = 1;
-	*ino = ip->i_number;
+	*ino = (INODETYPE)ip->i_number;
 	*nl = (long)ip->inode.iso_links;
-	*sz = ip->i_size;
+	*sz = (SZOFFTYPE)ip->i_size;
 # else	/* FREEBSDV>=2000 */
 	if (!v->v_data
 	||  kread((KA_T)v->v_data, (char *)&i, sizeof(i)))
 	    return(1);
 
 # if	FREEBSDV>=4000
-	if (i.i_dev && !kread((KA_T)i.i_dev, (char *)&udev, sizeof(udev))) {
+#  if	defined(HAS_NO_ISO_DEV)
+	if (i.i_mnt && !kread((KA_T)i.i_mnt, (char *)&im, sizeof(im))
+	&&  im.im_dev && !kread((KA_T)im.im_dev, (char *)&udev, sizeof(udev)))
+#  else	/* !defined(HAS_NO_ISO_DEV) */
+	if (i.i_dev && !kread((KA_T)i.i_dev, (char *)&udev, sizeof(udev)))
+#  endif	/* defined(HAS_NO_ISO_DEV) */
+
+	{
+
+# if	defined(HAS_NO_SI_UDEV)
+	    *d = Dev2Udev(&udev);
+# else	/* !defined(HAS_NO_SI_UDEV) */
 	    *d = udev.si_udev;
+# endif	/* defined(HAS_NO_SI_UDEV) */
+
 	    *dd = 1;
 	}
 # else	/* FREEBSDV<4000 */
@@ -112,9 +134,9 @@ read_iso_node(v, d, dd, ino, nl, sz)
 	*dd = 1;
 # endif	/* FREEBSDV>=4000 */
 
-	*ino = i.i_number;
+	*ino = (INODETYPE)i.i_number;
 	*nl = (long)i.inode.iso_links;
-	*sz = i.i_size;
+	*sz = (SZOFFTYPE)i.i_size;
 # endif	/* FREEBSDV<2000 */
 
 	return(0);
