@@ -33,21 +33,34 @@
 
 
 /*
- * $Id: dproto.h,v 1.7 99/11/28 06:40:29 abe Exp $
+ * $Id: dproto.h,v 1.13 2008/10/21 16:16:06 abe Exp $
  */
 
 
-#if	!defined(N_UNIXV)
+#if	FREEBSDV>=5000 && defined(HAS_NO_SI_UDEV)
+# if	defined(HAS_CONF_MINOR)
+_PROTOTYPE(extern dev_t Dev2Udev,(KA_T c));
+# else	/* !defined(HAS_CONF_MINOR) */
+_PROTOTYPE(extern dev_t Dev2Udev,(struct cdev *c));
+# endif	/* defined(HAS_CONF_MINOR) */
+#endif	/* FREEBSDV>=5000 && defined(HAS_NO_SI_UDEV) */
+
+#if	!defined(N_UNIX)
 _PROTOTYPE(extern char *get_nlist_path,(int ap));
-#endif	/* !defined(N_UNIXV) */
+#endif	/* !defined(N_UNIX) */
 
 _PROTOTYPE(extern int is_file_named,(char *p, int cd));
+_PROTOTYPE(extern void process_socket,(KA_T sa));
 _PROTOTYPE(extern struct l_vfs *readvfs,(KA_T vm));
 
-#if	FREEBSDV>=220
+#if	defined(HASKQUEUE)
+_PROTOTYPE(extern void process_kqueue,(KA_T ka));
+#endif	/* defined(HASKQUEUE) */
+
+#if	FREEBSDV>=2020
 _PROTOTYPE(extern void process_pipe,(KA_T pa));
-#endif	/* FREEBSDV>=220 */
+#endif	/* FREEBSDV>=2020 */
 
 #if	defined(HAS9660FS)
-_PROTOTYPE(extern int read_iso_node,(struct vnode *v, dev_t *d, int *dd, unsigned long *ino, long *nl, unsigned long *sz));
+_PROTOTYPE(extern int read_iso_node,(struct vnode *v, dev_t *d, int *dd, INODETYPE *ino, long *nl, SZOFFTYPE *sz));
 #endif	/* defined(HAS9660FS) */
