@@ -32,7 +32,7 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dsock.c,v 1.25 2004/07/06 19:18:57 abe Exp $";
+static char *rcsid = "$Id: dsock.c,v 1.27 2008/10/21 16:16:06 abe Exp $";
 #endif
 
 
@@ -50,6 +50,78 @@ static char *rcsid = "$Id: dsock.c,v 1.25 2004/07/06 19:18:57 abe Exp $";
 #endif	/* defined(HASIPv6) */
 
 
+/*
+ * Local function prototypes
+ */
+
+_PROTOTYPE(static int ckstate,(KA_T ta, struct tcpcb *t, int fam));
+
+
+/*
+ * ckstate() -- read TCP control block and check TCP state for inclusion
+ *		or exclusion
+ * return: -1 == no TCP CB available
+ *	    0 == TCP DB available; continue processing file
+ *	    1 == stop processing file
+ */
+
+static int
+ckstate(ta, t, fam)
+	KA_T ta;			/* TCP control block address */
+	struct tcpcb *t;		/* TCP control block receptor */
+	int fam;			/* protocol family */
+{
+	int tsnx;
+/*
+ * Read TCP control block.
+ */
+	if (kread(ta, (char *)t, sizeof(struct tcpcb)))
+	    return(-1);
+	if (TcpStXn || TcpStIn) {
+
+	/*
+	 * If there are TCP state inclusions or exclusions, check them.
+	 */
+	    tsnx = (int)t->t_state + TcpStOff;
+	    if (TcpStXn) {
+		if (TcpStX[tsnx]) {
+		    Lf->sf &= ~SELNET;
+		    Lf->sf |= SELEXCLF;
+		    return(1);
+		}
+	    }
+	    if (TcpStIn) {
+		if (TcpStI[tsnx]) {
+		    TcpStI[tsnx] = 2;
+		    Lf->sf |= SELNET;
+		} else {
+		    Lf->sf &= ~SELNET;
+		    Lf->sf |= SELEXCLF;
+		    return(1);
+		}
+	    }
+	}
+	if (!(Lf->sf & SELNET) && !TcpStIn) {
+
+	/*
+	 * See if this TCP file should be selected.
+	 */
+	    if (Fnet) {
+		if (!FnetTy
+		||  (FnetTy == 4) && (fam == AF_INET)
+
+#if	defined(HASIPv6)
+		||  (FnetTy == 6) && (fam == AF_INET6)
+#endif 	/* defined(HASIPv6) */
+
+		) {
+		    Lf->sf |= SELNET;
+		}
+	    }
+	}
+	return(0);
+}
+
 
 /*
  * process_socket() - process socket
@@ -68,7 +140,7 @@ process_socket(sa)
 	struct protosw p;
 	struct socket s;
 	struct tcpcb t;
-	KA_T ta = (KA_T)NULL;
+	int ts = -1;
 	struct unpcb uc, unp;
 	struct sockaddr_un *ua = NULL;
 	struct sockaddr_un un;
@@ -139,12 +211,13 @@ process_socket(sa)
 #if	defined(HASSOOPT)
 	Lf->lts.ltm = (unsigned int)s.so_linger;
 	Lf->lts.opt = (unsigned int)s.so_options;
+	Lf->lts.pqlen = (unsigned int)s.so_incqlen;
 	Lf->lts.qlen = (unsigned int)s.so_qlen;
 	Lf->lts.qlim = (unsigned int)s.so_qlimit;
 	Lf->lts.rbsz = (unsigned long)s.so_rcv.sb_mbmax;
 	Lf->lts.sbsz = (unsigned long)s.so_snd.sb_mbmax;
-	Lf->lts.qlens = Lf->lts.qlims = Lf->lts.rbszs
-		      = Lf->lts.sbszs = (unsigned char)1;
+	Lf->lts.pqlens = Lf->lts.qlens = Lf->lts.qlims = Lf->lts.rbszs
+		       = Lf->lts.sbszs = (unsigned char)1;
 #endif	/* defined(HASSOOPT) */
 
 #if	defined(HASSOSTATE)
@@ -175,9 +248,11 @@ process_socket(sa)
 #if	defined(HASIPv6)
 		||  ((FnetTy == 6) && (fam == AF_INET6))
 #endif	/* defined(HASIPv6) */
-		)
 
-		    Lf->sf |= SELNET;
+		) {
+		    if (!TcpStIn && !UdpStIn)
+			Lf->sf |= SELNET;
+		}
 	    }
 	    printiproto(p.pr_protocol);
 
@@ -204,11 +279,15 @@ process_socket(sa)
 	    /*
 	     * Save IPv6 address information.
 	     */
+		if (p.pr_protocol == IPPROTO_TCP) {
+		    if (in6p.in6p_ppcb) {
+			if ((ts = ckstate((KA_T)in6p.in6p_ppcb, &t, fam)) == 1)
+			    return;
+		    }
+		}
 		enter_dev_ch(print_kptr((KA_T)(in6p.in6p_ppcb ? in6p.in6p_ppcb
 							      : s.so_pcb),
 					       (char *)NULL, 0));
-		if (p.pr_protocol == IPPROTO_TCP)
-		    ta = (KA_T)in6p.in6p_ppcb;
 	        la = (unsigned char *)&in6p.in6p_laddr;
 	        lp = (int)ntohs(in6p.in6p_lport);
 		if (!IN6_IS_ADDR_UNSPECIFIED(&in6p.in6p_faddr)
@@ -251,11 +330,15 @@ process_socket(sa)
 		    enter_nm(Namech);
 		    return;
 		}
+		if (p.pr_protocol == IPPROTO_TCP) {
+		    if (inp.inp_ppcb) {
+			if ((ts = ckstate((KA_T)inp.inp_ppcb, &t, fam)) == 1)
+			    return;
+		    }
+		}
 		enter_dev_ch(print_kptr((KA_T)(inp.inp_ppcb ? inp.inp_ppcb
 							    : s.so_pcb),
 					       (char *)NULL, 0));
-		if (p.pr_protocol == IPPROTO_TCP)
-		    ta = (KA_T)inp.inp_ppcb;
 		lp = (int)ntohs(inp.inp_lport);
 		if (fam == AF_INET) {
 
@@ -305,7 +388,7 @@ process_socket(sa)
  	 */
 	    if (fa || la)
 		(void) ent_inaddr(la, lp, fa, fp, fam);
-	    if (ta && !kread(ta, (char *)&t, sizeof(t))) {
+	    if (ts == 0) {
 		Lf->lts.type = 0;
 		Lf->lts.state.i = (int)t.t_state;
 
