@@ -32,7 +32,7 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1997 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dsock.c,v 1.30 2005/08/29 10:23:33 abe Exp $";
+static char *rcsid = "$Id: dsock.c,v 1.33 2008/10/21 16:17:21 abe Exp $";
 #endif
 
 
@@ -43,8 +43,11 @@ static char *rcsid = "$Id: dsock.c,v 1.30 2005/08/29 10:23:33 abe Exp $";
  * Local definitions
  */
 
-#define	INOBUCKS	128		/* inode hash bucket count */
+#define	INOBUCKS	128		/* inode hash bucket count -- must be
+					 * a power of two */
 #define INOHASH(ino)	((int)((ino * 31415) >> 3) & (INOBUCKS - 1))
+#define TCPUDPHASH(ino)	((int)((ino * 31415) >> 3) & (TcpUdp_bucks - 1))
+#define TCPUDP6HASH(ino) ((int)((ino * 31415) >> 3) & (TcpUdp6_bucks - 1))
 
 
 /*
@@ -71,6 +74,13 @@ struct ipxsin {				/* IPX socket information */
 	struct ipxsin *next;
 };
 
+struct packin {				/* packet information */
+	INODETYPE inode;
+	int ty;				/* socket type */
+	int pr;				/* protocol */
+	struct packin *next;
+};
+
 struct rawsin {				/* raw socket information */
 	INODETYPE inode;
 	char *la;			/* local address */
@@ -88,7 +98,7 @@ struct tcp_udp {			/* IPv4 TCP and UDP socket
 	unsigned long faddr, laddr;	/* foreign & local IPv6 addresses */
 	int fport, lport;		/* foreign & local ports */
 	unsigned long txq, rxq;		/* transmit & receive queue values */
-	int proto;			/* 0 = TCP, 1 = UDP */
+	int proto;			/* 0 = TCP, 1 = UDP, 2 = UDPLITE */
 	int state;			/* protocol state */
 	struct tcp_udp *next;
 };
@@ -100,7 +110,7 @@ struct tcp_udp6 {			/* IPv6 TCP and UDP socket
 	struct in6_addr faddr, laddr;	/* foreign and local IPv6 addresses */
 	int fport, lport;		/* foreign & local ports */
 	unsigned long txq, rxq;		/* transmit & receive queue values */
-	int proto;			/* 0 = TCP, 1 = UDP */
+	int proto;			/* 0 = TCP, 1 = UDP, 2 = UDPLITE */
 	int state;			/* protocol state */
 	struct tcp_udp6 *next;
 };
@@ -136,28 +146,288 @@ static char *ax25st[] = {
 static char *Ipxpath = (char *)NULL;	/* path to IPX /proc information */
 static struct ipxsin **Ipxsin = (struct ipxsin **)NULL;
 					/* IPX socket info, hashed by inode */
+static struct packin **Packin = (struct packin **)NULL;
+					/* packet info, hashed by inode */
+static char *Packpath = (char *)NULL;	/* path to packer /proc information */
+struct packpr {				/* packet protocol conversions */
+	int pr;				/* protocol number */
+	char *nm;			/* protocol name */
+} static Packpr[] = {
+
+#if	defined(ETH_P_LOOP)
+	{ ETH_P_LOOP,		"LOOP" },
+#endif	/* defined(ETH_P_LOOP) */
+
+#if	defined(ETH_P_PUP)
+	{ ETH_P_PUP,		"PUP" },
+#endif	/* defined(ETH_P_PUP) */
+
+#if	defined(ETH_P_PUPAT)
+	{ ETH_P_PUPAT,		"PUPAT" },
+#endif	/* defined(ETH_P_PUPAT) */
+
+#if	defined(ETH_P_IP)
+	{ ETH_P_IP,		"IP" },
+#endif	/* defined(ETH_P_IP) */
+
+#if	defined(ETH_P_X25)
+	{ ETH_P_X25,		"X25" },
+#endif	/* defined(ETH_P_X25) */
+
+#if	defined(ETH_P_ARP)
+	{ ETH_P_ARP,		"ARP" },
+#endif	/* defined(ETH_P_ARP) */
+
+#if	defined(ETH_P_BPQ)
+	{ ETH_P_BPQ,		"BPQ" },
+#endif	/* defined(ETH_P_BPQ) */
+
+#if	defined(ETH_P_IEEEPUP)
+	{ ETH_P_IEEEPUP,	"I3EPUP" },
+#endif	/* defined(ETH_P_IEEEPUP) */
+
+#if	defined(ETH_P_IEEEPUPAT)
+	{ ETH_P_IEEEPUPAT,	"I3EPUPA" },
+#endif	/* defined(ETH_P_IEEEPUPAT) */
+
+#if	defined(ETH_P_DEC)
+	{ ETH_P_DEC,		"DEC" },
+#endif	/* defined(ETH_P_DEC) */
+
+#if	defined(ETH_P_DNA_DL)
+	{ ETH_P_DNA_DL,		"DNA_DL" },
+#endif	/* defined(ETH_P_DNA_DL) */
+
+#if	defined(ETH_P_DNA_RC)
+	{ ETH_P_DNA_RC,		"DNA_RC" },
+#endif	/* defined(ETH_P_DNA_RC) */
+
+#if	defined(ETH_P_DNA_RT)
+	{ ETH_P_DNA_RT,		"DNA_RT" },
+#endif	/* defined(ETH_P_DNA_RT) */
+
+#if	defined(ETH_P_LAT)
+	{ ETH_P_LAT,		"LAT" },
+#endif	/* defined(ETH_P_LAT) */
+
+#if	defined(ETH_P_DIAG)
+	{ ETH_P_DIAG,		"DIAG" },
+#endif	/* defined(ETH_P_DIAG) */
+
+#if	defined(ETH_P_CUST)
+	{ ETH_P_CUST,		"CUST" },
+#endif	/* defined(ETH_P_CUST) */
+
+#if	defined(ETH_P_SCA)
+	{ ETH_P_SCA,		"SCA" },
+#endif	/* defined(ETH_P_SCA) */
+
+#if	defined(ETH_P_RARP)
+	{ ETH_P_RARP,		"RARP" },
+#endif	/* defined(ETH_P_RARP) */
+
+#if	defined(ETH_P_ATALK)
+	{ ETH_P_ATALK,		"ATALK" },
+#endif	/* defined(ETH_P_ATALK) */
+
+#if	defined(ETH_P_AARP)
+	{ ETH_P_AARP,		"AARP" },
+#endif	/* defined(ETH_P_AARP) */
+
+#if	defined(ETH_P_8021Q)
+	{ ETH_P_8021Q,		"8021Q" },
+#endif	/* defined(ETH_P_8021Q) */
+
+#if	defined(ETH_P_IPX)
+	{ ETH_P_IPX,		"IPX" },
+#endif	/* defined(ETH_P_IPX) */
+
+#if	defined(ETH_P_IPV6)
+	{ ETH_P_IPV6,		"IPV6" },
+#endif	/* defined(ETH_P_IPV6) */
+
+#if	defined(ETH_P_SLOW)
+	{ ETH_P_SLOW,		"SLOW" },
+#endif	/* defined(ETH_P_SLOW) */
+	
+#if	defined(ETH_P_WCCP)
+	{ ETH_P_WCCP,		"WCCP" },
+#endif	/* defined(ETH_P_WCCP) */
+
+#if	defined(ETH_P_PPP_DISC)
+	{ ETH_P_PPP_DISC,	"PPP_DIS" },
+#endif	/* defined(ETH_P_PPP_DISC) */
+
+#if	defined(ETH_P_PPP_SES)
+	{ ETH_P_PPP_SES,	"PPP_SES" },
+#endif	/* defined(ETH_P_PPP_SES) */
+
+#if	defined(ETH_P_MPLS_UC)
+	{ ETH_P_MPLS_UC,	"MPLS_UC" },
+#endif	/* defined(ETH_P_MPLS_UC) */
+
+#if	defined(ETH_P_MPLS_MC)
+	{ ETH_P_MPLS_MC,	"MPLS_MC" },
+#endif	/* defined(ETH_P_MPLS_MC) */
+
+#if	defined(ETH_P_ATMMPOA)
+	{ ETH_P_ATMMPOA,	"ATMMPOA" },
+#endif	/* defined(ETH_P_ATMMPOA) */
+
+#if	defined(ETH_P_MPLS_MC)
+	{ ETH_P_MPLS_MC,	"MPLS_MC" },
+#endif	/* defined(ETH_P_MPLS_MC) */
+
+#if	defined(ETH_P_ATMFATE)
+	{ ETH_P_ATMFATE,	"ATMFATE" },
+#endif	/* defined(ETH_P_ATMFATE) */
+
+#if	defined(ETH_P_AOE)
+	{ ETH_P_AOE,		"AOE" },
+#endif	/* defined(ETH_P_AOE) */
+
+#if	defined(ETH_P_TIPC)
+	{ ETH_P_TIPC,		"TIPC" },
+#endif	/* defined(ETH_P_TIPC) */
+
+#if	defined(ETH_P_802_3)
+	{ ETH_P_802_3,		"802.3" },
+#endif	/* defined(ETH_P_802_3) */
+
+#if	defined(ETH_P_AX25)
+	{ ETH_P_AX25,		"AX25" },
+#endif	/* defined(ETH_P_AX25) */
+
+#if	defined(ETH_P_ALL)
+	{ ETH_P_ALL,		"ALL" },
+#endif	/* defined(ETH_P_ALL) */
+
+#if	defined(ETH_P_802_2)
+	{ ETH_P_802_2,		"802.2" },
+#endif	/* defined(ETH_P_802_2) */
+
+#if	defined(ETH_P_SNAP)
+	{ ETH_P_SNAP,		"SNAP" },
+#endif	/* defined(ETH_P_SNAP) */
+
+#if	defined(ETH_P_DDCMP)
+	{ ETH_P_DDCMP,		"DDCMP" },
+#endif	/* defined(ETH_P_DDCMP) */
+
+#if	defined(ETH_P_WAN_PPP)
+	{ ETH_P_WAN_PPP,	"WAN_PPP" },
+#endif	/* defined(ETH_P_WAN_PPP) */
+
+#if	defined(ETH_P_PPP_MP)
+	{ ETH_P_PPP_MP,		"PPP MP" },
+#endif	/* defined(ETH_P_PPP_MP) */
+
+#if	defined(ETH_P_LOCALTALK)
+	{ ETH_P_LOCALTALK,	"LCLTALK" },
+#endif	/* defined(ETH_P_LOCALTALK) */
+
+#if	defined(ETH_P_PPPTALK)
+	{ ETH_P_PPPTALK,	"PPPTALK" },
+#endif	/* defined(ETH_P_PPPTALK) */
+
+#if	defined(ETH_P_TR_802_2)
+	{ ETH_P_TR_802_2,	"802.2" },
+#endif	/* defined(ETH_P_TR_802_2) */
+
+#if	defined(ETH_P_MOBITEX)
+	{ ETH_P_MOBITEX,	"MOBITEX" },
+#endif	/* defined(ETH_P_MOBITEX) */
+
+#if	defined(ETH_P_CONTROL)
+	{ ETH_P_CONTROL,	"CONTROL" },
+#endif	/* defined(ETH_P_CONTROL) */
+
+#if	defined(ETH_P_IRDA)
+	{ ETH_P_IRDA,		"IRDA" },
+#endif	/* defined(ETH_P_IRDA) */
+
+#if	defined(ETH_P_ECONET)
+	{ ETH_P_ECONET,		"ECONET" },
+#endif	/* defined(ETH_P_ECONET) */
+
+#if	defined(ETH_P_HDLC)
+	{ ETH_P_HDLC,		"HDLC" },
+#endif	/* defined(ETH_P_HDLC) */
+
+#if	defined(ETH_P_ARCNET)
+	{ ETH_P_ARCNET,		"ARCNET" },
+#endif	/* defined(ETH_P_ARCNET) */
+
+};
+#define NPACKPR	(sizeof(Packpr) / sizeof(struct packpr))
+struct packty {				/* packet socket type conversions */
+	int ty;				/* type number */
+	char *nm;			/* type name */
+} static Packty[] = {
+
+#if	defined(SOCK_STREAM)
+	{ SOCK_STREAM,		"STREAM" },
+#endif	/* defined(SOCK_STREAM) */
+
+#if	defined(SOCK_DGRAM)
+	{ SOCK_DGRAM,		"DGRAM" },
+#endif	/* defined(SOCK_DGRAM) */
+
+#if	defined(SOCK_RAW)
+	{ SOCK_RAW,		"RAW" },
+#endif	/* defined(SOCK_RAW) */
+
+#if	defined(SOCK_RDM)
+	{ SOCK_RDM,		"RDM" },
+#endif	/* defined(SOCK_RDM) */
+
+#if	defined(SOCK_SEQPACKET)
+	{ SOCK_SEQPACKET,	"SEQPACKET" },
+#endif	/* defined(SOCK_SEQPACKET) */
+
+#if	defined(SOCK_PACKET)
+	{ SOCK_PACKET,		"PACKET" },
+#endif	/* defined(SOCK_PACKET) */
+};
+#define NPACKTY	(sizeof(Packty) / sizeof(struct packty))
 static char *Rawpath = (char *)NULL;	/* path to raw socket /proc
 					 * information */
 static struct rawsin **Rawsin = (struct rawsin **)NULL;
 					/* raw socket info, hashed by inode */
+static char *SockStatPath = (char *)NULL;
+					/* path to /proc/net socket status */
 static char *TCPpath = (char *)NULL;	/* path to TCP /proc information */
 static struct tcp_udp **TcpUdp = (struct tcp_udp **)NULL;
 					/* IPv4 TCP & UDP info, hashed by
 					 * inode */
+static int TcpUdp_bucks = 0;		/* dynamically sized hash bucket
+					 * count for TCP and UDP -- will
+					 * be a power of two */
 
 #if	defined(HASIPv6)
 static char *Raw6path = (char *)NULL;	/* path to raw IPv6 /proc information */
 static struct rawsin **Rawsin6 = (struct rawsin **)NULL;
 					/* IPv6 raw socket info, hashed by
 					 * inode */
+static char *SockStatPath6 = (char *)NULL;
+					/* path to /proc/net IPv6 socket
+					 * status */
 static char *TCP6path = (char *)NULL;	/* path to IPv6 TCP /proc information */
 static struct tcp_udp6 **TcpUdp6 = (struct tcp_udp6 **)NULL;
 					/* IPv6 TCP & UDP info, hashed by
 					 * inode */
+static int TcpUdp6_bucks = 0;		/* dynamically sized hash bucket
+					 * count for IPv6 TCP and UDP -- will
+					 * be a power of two */
 static char *UDP6path = (char *)NULL;	/* path to IPv6 UDP /proc information */
+static char *UDP6LITEpath = (char *)NULL;
+					/* path to IPv6 UDPLITE /proc
+					 * information */
 #endif	/* defined(HASIPv6) */
 
 static char *UDPpath = (char *)NULL;	/* path to UDP /proc information */
+static char *UDPLITEpath = (char *)NULL;
+					/* path to UDPLITE /proc information */
 static char *UNIXpath = (char *)NULL;	/* path to UNIX /proc information */
 static struct uxsin **Uxsin = (struct uxsin **)NULL;
 					/* UNIX socket info, hashed by inode */
@@ -169,11 +439,13 @@ static struct uxsin **Uxsin = (struct uxsin **)NULL;
 
 _PROTOTYPE(static struct ax25sin *check_ax25,(INODETYPE i));
 _PROTOTYPE(static struct ipxsin *check_ipx,(INODETYPE i));
+_PROTOTYPE(static struct packin *check_pack,(INODETYPE i));
 _PROTOTYPE(static struct rawsin *check_raw,(INODETYPE i));
 _PROTOTYPE(static struct tcp_udp *check_tcpudp,(INODETYPE i, char **p));
 _PROTOTYPE(static struct uxsin *check_unix,(INODETYPE i));
 _PROTOTYPE(static void get_ax25,(char *p));
 _PROTOTYPE(static void get_ipx,(char *p));
+_PROTOTYPE(static void get_pack,(char *p));
 _PROTOTYPE(static void get_raw,(char *p));
 _PROTOTYPE(static void get_tcpudp,(char *p, int pr, int clr));
 _PROTOTYPE(static void get_unix,(char *p));
@@ -189,6 +461,30 @@ _PROTOTYPE(static int net6a2in6,(char *as, struct in6_addr *ad));
 #endif	/* defined(HASIPv6) */
 
 
+/*
+ * build_IPstates() -- build the TCP and UDP state tables
+ */
+
+void
+build_IPstates()
+{
+	if (!TcpSt) {
+	    (void) enter_IPstate("TCP", "ESTABLISHED", TCP_ESTABLISHED);
+	    (void) enter_IPstate("TCP", "SYN_SENT", TCP_SYN_SENT);
+	    (void) enter_IPstate("TCP", "SYN_RECV", TCP_SYN_RECV);
+	    (void) enter_IPstate("TCP", "FIN_WAIT1", TCP_FIN_WAIT1);
+	    (void) enter_IPstate("TCP", "FIN_WAIT2", TCP_FIN_WAIT2);
+	    (void) enter_IPstate("TCP", "TIME_WAIT", TCP_TIME_WAIT);
+	    (void) enter_IPstate("TCP", "CLOSE", TCP_CLOSE);
+	    (void) enter_IPstate("TCP", "CLOSE_WAIT", TCP_CLOSE_WAIT);
+	    (void) enter_IPstate("TCP", "LAST_ACK", TCP_LAST_ACK);
+	    (void) enter_IPstate("TCP", "LISTEN", TCP_LISTEN);
+	    (void) enter_IPstate("TCP", "CLOSING", TCP_CLOSING);
+	    (void) enter_IPstate("TCP", "CLOSED", 0);
+	    (void) enter_IPstate("TCP", (char *)NULL, 0);
+	}
+}
+
 
 /*
  * check_ax25() - check for AX25 socket file
@@ -230,6 +526,27 @@ check_ipx(i)
 }
 
 
+/*
+ * check_pack() - check for packet file
+ */
+
+static struct packin *
+check_pack(i)
+	INODETYPE i;			/* packet file's inode number */
+{
+	int h;
+	struct packin *pp;
+
+	h = INOHASH(i);
+	for (pp = Packin[h]; pp; pp = pp->next) {
+	    if (i == pp->inode)
+		return(pp);
+	}
+	return((struct packin *)NULL);
+}
+
+
+
 /*
  * check_raw() - check for raw socket file
  */
@@ -262,10 +579,22 @@ check_tcpudp(i, p)
 	int h;
 	struct tcp_udp *tp;
 
-	h = INOHASH(i);
+	h = TCPUDPHASH(i);
 	for (tp = TcpUdp[h]; tp; tp = tp->next) {
 	    if (i == tp->inode) {
-		*p = tp->proto ? "UDP" : "TCP";
+		switch (tp->proto) {
+		case 0:
+		    *p = "TCP";
+		    break;
+		case 1:
+		    *p = "UDP";
+		    break;
+		case 2:
+		    *p = "UDPLITE";
+		    break;
+		default:
+		   *p = "unknown";
+		}
 		return(tp);
 	    }
 	}
@@ -306,10 +635,22 @@ check_tcpudp6(i, p)
 	int h;
 	struct tcp_udp6 *tp6;
 
-	h = INOHASH(i);
+	h = TCPUDP6HASH(i);
 	for (tp6 = TcpUdp6[h]; tp6; tp6 = tp6->next) {
 	    if (i == tp6->inode) {
-		*p = tp6->proto ? "UDP" : "TCP";
+		switch (tp6->proto) {
+		case 0:
+		    *p = "TCP";
+		    break;
+		case 1:
+		    *p = "UDP";
+		    break;
+		case 2:
+		    *p = "UDPLITE";
+		    break;
+		default:
+		   *p = "unknown";
+		}
 		return(tp6);
 	    }
 	}
@@ -350,9 +691,12 @@ get_ax25(p)
 	FILE *as;
 	char buf[MAXPATHLEN], *da, *dev_ch, *ep, **fp, *sa;
 	int h, nf;
-	unsigned long inode, rq, sq, state;
+	INODETYPE inode;
+	unsigned long rq, sq, state;
 	MALLOC_S len;
 	unsigned char rqs, sqs;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 /*
  * Do second time cleanup or first time setup.
  */
@@ -381,10 +725,10 @@ get_ax25(p)
 	    }
 	}
 /*
- * Open and read the /proc/net/ax25 file.  Store AX25 socket info in the
- * AX25sin[] hash buckets.
+ * Open the /proc/net/ax25 file, assign a page size buffer to the stream,
+ * and read it.  Store AX25 socket info in the AX25sin[] hash buckets.
  */
-	if (!(as = fopen(p, "r")))
+	if (!(as = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	while (fgets(buf, sizeof(buf) - 1, as)) {
 	    if ((nf = get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0)) < 24)
@@ -410,12 +754,12 @@ get_ax25(p)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[23] || !*fp[23]
-	    ||  (inode = strtoul(fp[23], &ep, 0)) == ULONG_MAX
+	    ||  (inode = strtoull(fp[23], &ep, 0)) == ULONG_MAX
 	    ||  !ep || *ep)
 		continue;
 	    h = INOHASH((INODETYPE)inode);
 	    for (ap = AX25sin[h]; ap; ap = ap->next) {
-		if ((INODETYPE)inode == ap->inode)
+		if (inode == ap->inode)
 		    break;
 	    }
 	    if (ap)
@@ -496,7 +840,7 @@ get_ax25(p)
 	    }
 	    ap->da = da;
 	    ap->dev_ch = dev_ch;
-	    ap->inode = (INODETYPE)inode;
+	    ap->inode = inode;
 	    ap->rq = rq;
 	    ap->rqs = rqs;
 	    ap->sa = sa;
@@ -521,9 +865,12 @@ get_ipx(p)
 	char buf[MAXPATHLEN], *ep, **fp, *la, *ra;
 	int fl = 1;
 	int h;
-	unsigned long inode, rxq, state, txq;
+	INODETYPE inode;
+	unsigned long rxq, state, txq;
 	struct ipxsin *ip, *np;
 	MALLOC_S len;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 	FILE *xs;
 /*
  * Do second time cleanup or first time setup.
@@ -551,10 +898,10 @@ get_ipx(p)
 	    }
 	}
 /*
- * Open and read the /proc/net/ipx file.  Store IPX socket info in the
- * Ipxsin[] hash buckets.
+ * Open the /proc/net/ipx file, assign a page size buffer to the stream,
+ * and read it.  Store IPX socket info in the Ipxsin[] hash buckets.
  */
-	if (!(xs = fopen(p, "r")))
+	if (!(xs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	while (fgets(buf, sizeof(buf) - 1, xs)) {
 	    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) < 7)
@@ -588,12 +935,12 @@ get_ipx(p)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[6] || !*fp[6]
-	    ||  (inode = strtoul(fp[6], &ep, 0)) == ULONG_MAX
+	    ||  (inode = strtoull(fp[6], &ep, 0)) == ULONG_MAX
 	    ||  !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = INOHASH(inode);
 	    for (ip = Ipxsin[h]; ip; ip = ip->next) {
-		if ((INODETYPE)inode == ip->inode)
+		if (inode == ip->inode)
 		    break;
 	    }
 	    if (ip)
@@ -653,7 +1000,7 @@ get_ipx(p)
 		    Pn, sizeof(struct ipxsin));
 		Exit(1);
 	    }
-	    ip->inode = (INODETYPE)inode;
+	    ip->inode = inode;
 	    ip->la = la;
 	    ip->ra = ra;
 	    ip->txq = txq;
@@ -666,6 +1013,118 @@ get_ipx(p)
 }
 
 
+/*
+ * get_pack() - get /proc/net/packet info
+ */
+
+static void
+get_pack(p)
+	char *p;			/* /proc/net/raw path */
+{
+	char buf[MAXPATHLEN], *ep, **fp;
+	int h, lc, ty;
+	INODETYPE inode;
+	struct packin *np, *pp;
+	unsigned long pr;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
+	FILE *xs;
+/*
+ * Do second time cleanup or first time setup.
+ */
+	if (Packin) {
+	    for (h = 0; h < INOBUCKS; h++) {
+		for (pp = Packin[h]; pp; pp = np) {
+		    np = pp->next;
+		    (void) free((FREE_P *)pp);
+		}
+		Packin[h] = (struct packin *)NULL;
+	    }
+	} else {
+	    Packin = (struct packin **)calloc(INOBUCKS,
+					      sizeof(struct packin *));
+	    if (!Packin) {
+		(void) fprintf(stderr,
+		    "%s: can't allocate %d packet hash pointer bytes\n",
+		    Pn, INOBUCKS * sizeof(struct packin *));
+		Exit(1);
+	    }
+	}
+/*
+ * Open the /proc/net/packet file, assign a page size buffer to its stream,
+ * and read the file.  Store packet info in the Packin[] hash buckets.
+ */
+	if (!(xs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
+	    return;
+	lc = 0;
+	while (fgets(buf, sizeof(buf) - 1, xs)) {
+	    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) < 9)
+		continue;
+	    lc++;
+	    if (lc == 1) {
+
+	    /*
+	     * Check the column labels in the first line.
+	     */
+		if (!fp[2]  || strcmp(fp[2], "Type")
+		||  !fp[3]  || strcmp(fp[3], "Proto")
+		||  !fp[8] || strcmp(fp[8], "Inode"))
+		{
+		    if (!Fwarn) {
+			(void) fprintf(stderr,
+			    "%s: WARNING: unsupported format: %s\n",
+			    Pn, p);
+		    }
+		    break;
+		}
+		continue;
+	    }
+	/*
+	 * Assemble the inode number and see if the inode is already
+	 * recorded.
+	 */
+	    ep = (char *)NULL;
+	    if (!fp[8] || !*fp[8]
+	    ||  (inode = strtoull(fp[8], &ep, 0)) == ULONG_MAX
+	    ||  !ep || *ep)
+		continue;
+	    h = INOHASH(inode);
+	    for (pp = Packin[h]; pp; pp = pp->next) {
+		if (inode == pp->inode)
+		    break;
+	    }
+	    if (pp)
+		continue;
+	/*
+	 * Save the socket type and protocol.
+	 */
+	    if (!fp[2] || !*fp[2] || (strlen(fp[2])) < 1)
+		continue;
+	    ty = atoi(fp[2]);
+	    ep = (char *)NULL;
+	    if (!fp[3] || !*fp[3] || (strlen(fp[3]) < 1)
+	    ||  ((pr = strtoul(fp[3], &ep, 16)) == ULONG_MAX) || !ep || *ep)
+		continue;
+	/*
+	 * Allocate space for a packin entry, fill it, and link it to its
+	 * hash bucket.
+	 */
+	    if (!(pp = (struct packin *)malloc(sizeof(struct packin)))) {
+		(void) fprintf(stderr,
+		    "%s: can't allocate %d byte packet structure\n",
+		    Pn, sizeof(struct packin));
+		Exit(1);
+	    }
+	    pp->inode = inode;
+	    pp->pr = (int)pr;
+	    pp->ty = ty;
+	    pp->next = Packin[h];
+	    Packin[h] = pp;
+	}
+	(void) fclose(xs);
+}
+
+
 /*
  * get_raw() - get /proc/net/raw info
  */
@@ -676,10 +1135,12 @@ get_raw(p)
 {
 	char buf[MAXPATHLEN], *ep, **fp, *la, *ra, *sp;
 	int h;
-	unsigned long inode;
+	INODETYPE inode;
 	int nf = 12;
 	struct rawsin *np, *rp;
 	MALLOC_S lal, ral, spl;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 	FILE *xs;
 /*
  * Do second time cleanup or first time setup.
@@ -707,10 +1168,10 @@ get_raw(p)
 	    }
 	}
 /*
- * Open and read the /proc/net/raw file.  Store raw socket info in the
- * Rawsin[] hash buckets.
+ * Open the /proc/net/raw file, assign a page size buffer to its stream,
+ * and read the file.  Store raw socket info in the Rawsin[] hash buckets.
  */
-	if (!(xs = fopen(p, "r")))
+	if (!(xs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	while (fgets(buf, sizeof(buf) - 1, xs)) {
 	    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) < nf)
@@ -741,12 +1202,12 @@ get_raw(p)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[9] || !*fp[9]
-	    ||  (inode = strtoul(fp[9], &ep, 0)) == ULONG_MAX
+	    ||  (inode = strtoull(fp[9], &ep, 0)) == ULONG_MAX
 	    ||  !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = INOHASH(inode);
 	    for (rp = Rawsin[h]; rp; rp = rp->next) {
-		if ((INODETYPE)inode == rp->inode)
+		if (inode == rp->inode)
 		    break;
 	    }
 	    if (rp)
@@ -800,7 +1261,7 @@ get_raw(p)
 		    Pn, sizeof(struct rawsin));
 		Exit(1);
 	    }
-	    rp->inode = (INODETYPE)inode;
+	    rp->inode = inode;
 	    rp->la = la;
 	    rp->lal = lal;
 	    rp->ra = ra;
@@ -815,26 +1276,30 @@ get_raw(p)
 
 
 /*
- * get_tcpudp() - get IPv4 TCP or UDP IPv4 net info
+ * get_tcpudp() - get IPv4 TCP, UDP or UDPLITE net info
  */
 
 static void
 get_tcpudp(p, pr, clr)
 	char *p;			/* /proc/net/{tcp,udp} path */
-	int pr;				/* protocol: 0 = TCP, 1 = UDP */
+	int pr;				/* protocol: 0 = TCP, 1 = UDP,
+					 *           2 = UDPLITE */
 	int clr;			/* 1 == clear the table */
 {
 	char buf[MAXPATHLEN], *ep, **fp;
-	unsigned long faddr, fport, inode, laddr, lport, rxq, state, txq;
-	int h, nf;
+	unsigned long faddr, fport, laddr, lport, rxq, state, txq;
 	FILE *fs;
+	int h, nf;
+	INODETYPE inode;
 	struct tcp_udp *np, *tp;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 /*
- * Delete previous table contents.  Allocate a table for the first time.
+ * Delete previous table contents.
  */
 	if (TcpUdp) {
 	    if (clr) {
-		for (h = 0; h < INOBUCKS; h++) {
+		for (h = 0; h < TcpUdp_bucks; h++) {
 		    for (tp = TcpUdp[h]; tp; tp = np) {
 			np = tp->next;
 			(void) free((FREE_P *)tp);
@@ -842,20 +1307,46 @@ get_tcpudp(p, pr, clr)
 		    TcpUdp[h] = (struct tcp_udp *)NULL;
 		}
 	    }
+/*
+ * If no hash buckets have been allocated, do so now.
+ */
 	} else {
-	    if (!(TcpUdp = (struct tcp_udp **)calloc(INOBUCKS,
+	
+	/*
+	 * Open the /proc/net/sockstat file and establish the hash bucket
+	 * count from its "sockets: used" line.
+	 */
+	    TcpUdp_bucks = INOBUCKS;
+	    if ((fs = fopen(SockStatPath, "r"))) {
+		while(fgets(buf, sizeof(buf) - 1, fs)) {
+		    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) != 3)
+			continue;
+		    if (!fp[0] || strcmp(fp[0], "sockets:")
+		    ||  !fp[1] || strcmp(fp[1], "used")
+		    ||  !fp[2] || !*fp[2])
+			continue;
+    		    if ((h = atoi(fp[2])) < 1)
+			h = INOBUCKS;
+		    while (TcpUdp_bucks < h)
+			TcpUdp_bucks *= 2;
+		    break;
+		}
+		(void) fclose(fs);
+	    }
+	    if (!(TcpUdp = (struct tcp_udp **)calloc(TcpUdp_bucks,
 						     sizeof(struct tcp_udp *))))
 	    {
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for TCP&UDP hash buckets\n",
-		    Pn, INOBUCKS * sizeof(struct tcp_udp *));
+		    Pn, (int)(TcpUdp_bucks * sizeof(struct tcp_udp *)));
 		Exit(1);
 	    }
 	}
 /*
- * Open and read the /proc/net file.
- */
-	if (!(fs = fopen(p, "r")))
+ * Open the /proc/net file, assign a page size buffer to the stream, and
+ * read it.
+ */ 
+	if (!(fs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	nf = 12;
 	while(fgets(buf, sizeof(buf) - 1, fs)) {
@@ -902,7 +1393,7 @@ get_tcpudp(p, pr, clr)
 	    ||  (fport = strtoul(fp[4], &ep, 16)) == ULONG_MAX || !ep || *ep)
 		continue;
 	/*
-	 * Get the state, queue sizes, and inode.
+	 * Get the state and queue sizes.
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[5] || !*fp[5]
@@ -921,11 +1412,11 @@ get_tcpudp(p, pr, clr)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[13] || !*fp[13]
-	    ||  (inode = strtoul(fp[13], &ep, 0)) == ULONG_MAX || !ep || *ep)
+	    ||  (inode = strtoull(fp[13], &ep, 0)) == ULONG_MAX || !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = TCPUDPHASH(inode);
 	    for (tp = TcpUdp[h]; tp; tp = tp->next) {
-		if (tp->inode == (INODETYPE)inode)
+		if (tp->inode == inode)
 		    break;
 	    }
 	    if (tp)
@@ -939,7 +1430,7 @@ get_tcpudp(p, pr, clr)
 		    Pn, sizeof(struct tcp_udp));
 		Exit(1);
 	    }
-	    tp->inode = (INODETYPE)inode;
+	    tp->inode = inode;
 	    tp->faddr = faddr;
 	    tp->fport = (int)(fport & 0xffff);
 	    tp->laddr = laddr;
@@ -966,10 +1457,12 @@ get_raw6(p)
 {
 	char buf[MAXPATHLEN], *ep, **fp, *la, *ra, *sp;
 	int h;
-	unsigned long inode;
+	INODETYPE inode;
 	int nf = 12;
 	struct rawsin *np, *rp;
 	MALLOC_S lal, ral, spl;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 	FILE *xs;
 /*
  * Do second time cleanup or first time setup.
@@ -997,10 +1490,10 @@ get_raw6(p)
 	    }
 	}
 /*
- * Open and read the /proc/net/raw6 file.  Store raw6 socket info in the
- * Rawsin6[] hash buckets.
+ * Open the /proc/net/raw6 file, assign a page size buffer to the stream,
+ * and read it.  Store raw6 socket info in the Rawsin6[] hash buckets.
  */
-	if (!(xs = fopen(p, "r")))
+	if (!(xs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	while (fgets(buf, sizeof(buf) - 1, xs)) {
 	    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) < nf)
@@ -1031,12 +1524,12 @@ get_raw6(p)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[9] || !*fp[9]
-	    ||  (inode = strtoul(fp[9], &ep, 0)) == ULONG_MAX
+	    ||  (inode = strtoull(fp[9], &ep, 0)) == ULONG_MAX
 	    ||  !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = INOHASH(inode);
 	    for (rp = Rawsin6[h]; rp; rp = rp->next) {
-		if ((INODETYPE)inode == rp->inode)
+		if (inode == rp->inode)
 		    break;
 	    }
 	    if (rp)
@@ -1090,7 +1583,7 @@ get_raw6(p)
 		    Pn, sizeof(struct rawsin));
 		Exit(1);
 	    }
-	    rp->inode = (INODETYPE)inode;
+	    rp->inode = inode;
 	    rp->la = la;
 	    rp->lal = lal;
 	    rp->ra = ra;
@@ -1105,7 +1598,7 @@ get_raw6(p)
 
 
 /*
- * get_tcpudp6() - get IPv6 TCP or UDP IPv4 net info
+ * get_tcpudp6() - get IPv6 TCP, UDP or UDPLITE net info
  */
 
 static void
@@ -1116,16 +1609,19 @@ get_tcpudp6(p, pr, clr)
 {
 	char buf[MAXPATHLEN], *ep, **fp;
 	struct in6_addr faddr, laddr;
-	unsigned long fport, inode, lport, rxq, state, txq;
-	int h, nf;
+	unsigned long fport, lport, rxq, state, txq;
 	FILE *fs;
+	int h, i, nf;
+	INODETYPE inode;
 	struct tcp_udp6 *np6, *tp6;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 /*
  * Delete previous table contents.  Allocate a table for the first time.
  */
 	if (TcpUdp6) {
 	    if (clr) {
-		for (h = 0; h < INOBUCKS; h++) {
+		for (h = 0; h < TcpUdp6_bucks; h++) {
 		    for (tp6 = TcpUdp6[h]; tp6; tp6 = np6) {
 			np6 = tp6->next;
 			(void) free((FREE_P *)tp6);
@@ -1134,19 +1630,55 @@ get_tcpudp6(p, pr, clr)
 		}
 	    }
 	} else {
-	    if (!(TcpUdp6 = (struct tcp_udp6 **)calloc(INOBUCKS,
+	
+	/*
+	 * Open the /proc/net/sockstat6 file and establish the hash bucket
+	 * count from its "TCP6: inuse" and "UDP6: inuse" lines.
+	 */
+	    TcpUdp6_bucks = INOBUCKS;
+	    h = i = nf = 0;
+	    if ((fs = fopen(SockStatPath6, "r"))) {
+		while(fgets(buf, sizeof(buf) - 1, fs)) {
+		    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) != 3)
+			continue;
+		    if (!fp[0]
+		    ||  !fp[1] || strcmp(fp[1], "inuse")
+		    ||  !fp[2] || !*fp[2])
+			continue;
+		    if (!strcmp(fp[0], "TCP6:")) {
+			nf |= 1;
+    			if ((h = atoi(fp[2])) < 1)
+			    h = INOBUCKS;
+			i += h;
+		    } else if (!strcmp(fp[0], "UDP6:")) {
+			nf |= 2;
+    			if ((h = atoi(fp[2])) < 1)
+			    h = INOBUCKS;
+			i += h;
+		    } else
+			continue;
+		    if (nf == 3) {
+			while (TcpUdp6_bucks < i)
+			    TcpUdp6_bucks *= 2;
+			break;
+		    }
+		}
+		(void) fclose(fs);
+	    }
+	    if (!(TcpUdp6 = (struct tcp_udp6 **)calloc(TcpUdp6_bucks,
 						sizeof(struct tcp_udp6 *))))
 	    {
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for TCP6&UDP6 hash buckets\n",
-		    Pn, INOBUCKS * sizeof(struct tcp_udp6 *));
+		    Pn, (int)(TcpUdp6_bucks * sizeof(struct tcp_udp6 *)));
 		Exit(1);
 	    }
 	}
 /*
- * Open and read the /proc/net file.
+ * Open the /proc/net file, assign a page size buffer to the stream,
+ * and read it.
  */
-	if (!(fs = fopen(p, "r")))
+	if (!(fs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	nf = 12;
 	while(fgets(buf, sizeof(buf) - 1, fs)) {
@@ -1189,7 +1721,7 @@ get_tcpudp6(p, pr, clr)
 	    ||  (fport = strtoul(fp[4], &ep, 16)) == ULONG_MAX || !ep || *ep)
 		continue;
 	/*
-	 * Get the state, queue sizes, and inode.
+	 * Get the state and queue sizes.
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[5] || !*fp[5]
@@ -1208,11 +1740,11 @@ get_tcpudp6(p, pr, clr)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[13] || !*fp[13]
-	    ||  (inode = strtoul(fp[13], &ep, 0)) == ULONG_MAX || !ep || *ep)
+	    ||  (inode = strtoull(fp[13], &ep, 0)) == ULONG_MAX || !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = TCPUDP6HASH(inode);
 	    for (tp6 = TcpUdp6[h]; tp6; tp6 = tp6->next) {
-		if (tp6->inode == (INODETYPE)inode)
+		if (tp6->inode == inode)
 		    break;
 	    }
 	    if (tp6)
@@ -1226,7 +1758,7 @@ get_tcpudp6(p, pr, clr)
 		    Pn, sizeof(struct tcp_udp6));
 		Exit(1);
 	    }
-	    tp6->inode = (INODETYPE)inode;
+	    tp6->inode = inode;
 	    tp6->faddr = faddr;
 	    tp6->fport = (int)(fport & 0xffff);
 	    tp6->laddr = laddr;
@@ -1254,10 +1786,12 @@ get_unix(p)
 	char buf[MAXPATHLEN], *ep, **fp, *path, *pcb;
 	int fl = 1;
 	int h, nf;
-	unsigned long inode;
+	INODETYPE inode;
 	MALLOC_S len;
 	struct uxsin *np, *up;
 	FILE *us;
+	static char *vbuf = (char *)NULL;
+	static size_t vsz = (size_t)0;
 /*
  * Do second time cleanup or first time setup.
  */
@@ -1267,6 +1801,8 @@ get_unix(p)
 		    np = up->next;
 		    if (up->path)
 			(void) free((FREE_P *)up->path);
+		    if (up->pcb)
+			(void) free((FREE_P *)up->pcb);
 		    (void) free((FREE_P *)up);
 		}
 		Uxsin[h] = (struct uxsin *)NULL;
@@ -1280,10 +1816,10 @@ get_unix(p)
 	    }
 	}
 /*
- * Open /proc/net/unix, read its contents, and add them to the Uxsin hash
- * buckets.
+ * Open the /proc/net/unix file, assign a page size buffer to the stream,
+ * read the file's contents, and add them to the Uxsin hash buckets.
  */
-	if (!(us = fopen(p, "r")))
+	if (!(us = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
 	while (fgets(buf, sizeof(buf) - 1, us)) {
 	    if ((nf = get_fields(buf, ":", &fp, (int *)NULL, 0)) < 7)
@@ -1319,11 +1855,11 @@ get_unix(p)
 	 */
 	    ep = (char *)NULL;
 	    if (!fp[6] || !*fp[6]
-	    ||  (inode = strtoul(fp[6], &ep, 0)) == ULONG_MAX || !ep || *ep)
+	    ||  (inode = strtoull(fp[6], &ep, 0)) == ULONG_MAX || !ep || *ep)
 		continue;
-	    h = INOHASH((INODETYPE)inode);
+	    h = INOHASH(inode);
 	    for (up = Uxsin[h]; up; up = up->next) {
-		if ((INODETYPE)inode == up->inode)
+		if (inode == up->inode)
 		    break;
 	    }
 	    if (up)
@@ -1362,7 +1898,7 @@ get_unix(p)
 		    Pn, sizeof(struct uxsin));
 		Exit(1);
 	    }
-	    up->inode = (INODETYPE)inode;
+	    up->inode = inode;
 	    up->pcb = pcb;
 	    up->sb_def = 0;
 	    if ((up->path = path) && (*path == '/')) {
@@ -1520,47 +2056,14 @@ print_tcptpi(nl)
 	int s;
 
 	if ((Ftcptpi & TCPTPI_STATE) && Lf->lts.type == 0) {
-	    switch ((s = Lf->lts.state.i)) {
-	    case TCP_ESTABLISHED:
-		cp = "ESTABLISHED";
-		break;
-	    case TCP_SYN_SENT:
-		cp = "SYN_SENT";
-		break;
-	    case TCP_SYN_RECV:
-		cp = "SYN_RECV";
-		break;
-	    case TCP_FIN_WAIT1:
-		cp = "FIN_WAIT1";
-		break;
-	    case TCP_FIN_WAIT2:
-		cp = "FIN_WAIT2";
-		break;
-	    case TCP_TIME_WAIT:
-		cp = "TIME_WAIT";
-		break;
-	    case TCP_CLOSE:
-		cp = "CLOSE";
-		break;
-	    case TCP_CLOSE_WAIT:
-		cp = "CLOSE_WAIT";
-		break;
-	    case TCP_LAST_ACK:
-		cp = "LAST_ACK";
-		break;
-	    case TCP_LISTEN:
-		cp = "LISTEN";
-		break;
-	    case TCP_CLOSING:
-		cp = "CLOSING";
-		break;
-	    case 0:
-		cp = "CLOSED";
-		break;
-	    default:
-		(void) snpf(buf, sizeof(buf), "UNKNOWN_TCP_STATE_%d", s);
+	    if (!TcpSt)
+		(void) build_IPstates();
+	    if ((s = Lf->lts.state.i + TcpStOff) < 0 || s >= TcpNstates) {
+		(void) snpf(buf, sizeof(buf), "UNKNOWN_TCP_STATE_%d",
+		    Lf->lts.state.i);
 		cp = buf;
-    	    }
+    	    } else
+		cp = TcpSt[s];
 	    if (cp) {
 		if (Ffield)
 		    (void) printf("%cST=%s%c", LSOF_FID_TCPTPI, cp, Terminator);
@@ -1656,14 +2159,15 @@ process_proc_sock(p, s, ss, l, lss)
 	int ss;				/* *s status -- i.e, SB_* values */
 	struct stat *l;			/* lstat() result for FD (NULL for
 					 * others) */
-	int lss;				/* *l status -- i.e, SB_* values */
+	int lss;			/* *l status -- i.e, SB_* values */
 {
 	struct ax25sin *ap;
 	char *cp, dev_ch[32], *path;
 	unsigned char *fa, *la;
 	struct in_addr fs, ls;
 	struct ipxsin *ip;
-	int len, nl;
+	int i, len, nl;
+	struct packin *pp;
 	char *pr;
 	struct rawsin *rp;
 	struct tcp_udp *tp;
@@ -1678,7 +2182,7 @@ process_proc_sock(p, s, ss, l, lss)
  * Enter offset, if possible.
  */
 	if (Foffset || !Fsize) {
-	    if (l && (lss & SB_SIZE)) {
+	    if (l && (lss & SB_SIZE) && OffType) {
 		Lf->off = (SZOFFTYPE)l->st_size;
 		Lf->off_def = 1;
 	    }
@@ -1735,7 +2239,8 @@ process_proc_sock(p, s, ss, l, lss)
 		Lf->dev_def = 1;
 	    }
 	    cp = Namech;
-	    nl = MAXPATHLEN - 2;
+	    nl = Namechl;
+	    *cp = '\0';
 	    if (ip->la && nl) {
 
 	    /*
@@ -1791,7 +2296,8 @@ process_proc_sock(p, s, ss, l, lss)
 		Lf->inp_ty = 1;
 	    }
 	    cp = Namech;
-	    nl = MAXPATHLEN - 2;
+	    nl = Namechl - 2;
+	    *cp = '\0';
 	    if (rp->la && rp->lal) {
 
 	    /*
@@ -1834,6 +2340,46 @@ process_proc_sock(p, s, ss, l, lss)
 		enter_nm(Namech);
 	    return;
 	}
+	if (Packpath) {
+	    (void) get_pack(Packpath);
+	    (void) free((FREE_P *)Packpath);
+	    Packpath = (char *)NULL;
+	}
+	if ((ss & SB_INO)
+	&&  (pp = check_pack((INODETYPE)s->st_ino))
+	) {
+
+	/*
+	 * The inode is connected to a packet /proc record.
+	 *
+	 * Set the type to "pack"and  store the socket type in the NAME
+	 * column.  Put the protocol name in the NODE column and the inode
+	 * number in the DEVICE column.
+	 */
+	    (void) snpf(Lf->type, sizeof(Lf->type), "pack");
+	    for (i = 0; i < NPACKTY; i++) {
+		if (Packty[i].ty == pp->ty) {
+		    (void) snpf(Namech, Namechl, "type=SOCK_%s",
+			Packty[i].nm);
+		    break;
+		}
+	    }
+	    for (i = 0; i < NPACKPR; i++) {
+		if (Packpr[i].pr == pp->pr)
+		    break;
+	    }
+	    cp = (i < NPACKPR) ? Packpr[i].nm : "unknown";
+	    (void) snpf(Lf->iproto, sizeof(Lf->iproto), "%.*s", IPROTOL-1, cp);
+	    Lf->inp_ty = 2;
+	    if (ss & SB_INO) {
+		(void) snpf(dev_ch, sizeof(dev_ch), InodeFmt_d,
+		    (INODETYPE)s->st_ino);
+		enter_dev_ch(dev_ch);
+	    }
+	    if (Namech[0])
+		enter_nm(Namech);
+	    return;
+	}
 	if (UNIXpath) {
 	    (void) get_unix(UNIXpath);
 	    (void) free((FREE_P *)UNIXpath);
@@ -2002,6 +2548,12 @@ process_proc_sock(p, s, ss, l, lss)
 	    (void) free((FREE_P *)UDP6path);
 	    UDP6path = (char *)NULL;
 	}
+	if (UDP6LITEpath) {
+	    if (!Fxopt)
+		(void) get_tcpudp6(UDP6LITEpath, 2, 0);
+	    (void) free((FREE_P *)UDP6LITEpath);
+	    UDP6LITEpath = (char *)NULL;
+	}
 	if (!Fxopt && (ss & SB_INO)
 	&&  (tp6 = check_tcpudp6((INODETYPE)s->st_ino, &pr))
 	) {
@@ -2014,6 +2566,33 @@ process_proc_sock(p, s, ss, l, lss)
 	 * and foreign IPv6 addresses; save the type and protocol; and
 	 * (optionally) save the queue sizes.
 	 */
+	    i = tp6->state + TcpStOff;
+	    if (TcpStXn) {
+
+	    /*
+	     * Check for state exclusion.
+	     */
+		if (i >= 0 && i < TcpNstates) {
+		    if (TcpStX[i]) {
+			Lf->sf |= SELEXCLF;
+			return;
+		    }
+		}
+	    }
+	    if (TcpStIn) {
+
+	    /*
+	     * Check for state inclusion.
+	     */
+		if (i >= 0 && i < TcpNstates) {
+		    if (TcpStI[i])
+			TcpStI[i] = 2;
+		    else {
+			Lf->sf |= SELEXCLF;
+			return;
+		   }
+		}
+	    }
 	    if (Fnet && (FnetTy != 4))
 		Lf->sf |= SELNET;
 	    (void) snpf(Lf->type, sizeof(Lf->type), "IPv6");
@@ -2067,6 +2646,12 @@ process_proc_sock(p, s, ss, l, lss)
 	    (void) free((FREE_P *)UDPpath);
 	    UDPpath = (char *)NULL;
 	}
+	if (UDPLITEpath) {
+	    if (!Fxopt)
+		(void) get_tcpudp(UDPLITEpath, 2, 0);
+	    (void) free((FREE_P *)UDPLITEpath);
+	    UDPLITEpath = (char *)NULL;
+	}
 	if (!Fxopt && (ss & SB_INO)
 	&&  (tp = check_tcpudp((INODETYPE)s->st_ino, &pr))
 	) {
@@ -2079,6 +2664,33 @@ process_proc_sock(p, s, ss, l, lss)
 	 * save the local and foreign IPv4 addresses; save the type and
 	 * protocol; and (optionally) save the queue sizes.
 	 */
+	    i = tp->state + TcpStOff;
+	    if (TcpStXn) {
+
+	    /*
+	     * Check for state exclusion.
+	     */
+		if (i >= 0 && i < TcpNstates) {
+		    if (TcpStX[i]) {
+			Lf->sf |= SELEXCLF;
+			return;
+		    }
+		}
+	    }
+	    if (TcpStIn) {
+
+	    /*
+	     * Check for state inclusion.
+	     */
+		if (i >= 0 && i < TcpNstates) {
+		    if (TcpStI[i])
+			TcpStI[i] = 2;
+		    else {
+			Lf->sf |= SELEXCLF;
+			return;
+		    }
+		}
+	    }
 	    if (Fnet && (FnetTy != 6))
 		Lf->sf |= SELNET;
 
@@ -2150,19 +2762,29 @@ set_net_paths(p, pl)
 	pathl = 0;
 	(void) make_proc_path(p, pl, &Ipxpath, &pathl, "ipx");
 	pathl = 0;
+	(void) make_proc_path(p, pl, &Packpath, &pathl, "packet");
+	pathl = 0;
 	(void) make_proc_path(p, pl, &Rawpath, &pathl, "raw");
 	pathl = 0;
+	(void) make_proc_path(p, pl, &SockStatPath, &pathl, "sockstat");
+	pathl = 0;
 	(void) make_proc_path(p, pl, &TCPpath, &pathl, "tcp");
 	pathl = 0;
 	(void) make_proc_path(p, pl, &UDPpath, &pathl, "udp");
+	pathl = 0;
+	(void) make_proc_path(p, pl, &UDPLITEpath, &pathl, "udplite");
 
 #if	defined(HASIPv6)
 	pathl = 0;
 	(void) make_proc_path(p, pl, &Raw6path, &pathl, "raw6");
 	pathl = 0;
+	(void) make_proc_path(p, pl, &SockStatPath6, &pathl, "sockstat6");
+	pathl = 0;
 	(void) make_proc_path(p, pl, &TCP6path, &pathl, "tcp6");
 	pathl = 0;
 	(void) make_proc_path(p, pl, &UDP6path, &pathl, "udp6");
+	pathl = 0;
+	(void) make_proc_path(p, pl, &UDP6LITEpath, &pathl, "udp6lite");
 #endif	/* defined(HASIPv6) */
 
 	pathl = 0;
