@@ -33,10 +33,14 @@
 
 
 /*
- * $Id: dproto.h,v 1.13 2001/07/05 12:26:21 abe Exp $
+ * $Id: dproto.h,v 1.20 2008/10/21 16:16:42 abe Exp $
  */
 
 
+#if	defined(HASVXFSUTIL)
+_PROTOTYPE(extern int access_vxfs_ioffsets,(void));
+#endif	/* defined(HASVXFSUTIL) */
+
 _PROTOTYPE(extern void completevfs,(struct l_vfs *vfs, dev_t *dev));
 _PROTOTYPE(extern int is_file_named,(char *p, int nt, enum vtype vt, int ps));
 _PROTOTYPE(extern struct l_vfs *readvfs,(KA_T ka, struct vfs *la, struct vnode *lv));
@@ -55,11 +59,27 @@ _PROTOTYPE(extern void clr_sect,(void));
 _PROTOTYPE(extern int rw_pseudo_sect,(int m));
 #endif	/* defined(HASDCACHE) */
 
+#if	defined(HASIPv6)
+_PROTOTYPE(extern struct hostent *gethostbyname2,(const char *nm, int proto));
+#endif	/* defined(HASIPv6) */
+
+#if	defined(HAS_V_PATH)
+_PROTOTYPE(extern int print_v_path,(struct lfile *lf));
+_PROTOTYPE(extern void read_v_path,(KA_T ka, char *rb, size_t rbl));
+#endif	/* defined(HAS_V_PATH) */
+
 #if	defined(HASVXFS)
 _PROTOTYPE(extern int read_vxnode,(KA_T va, struct vnode *v, struct l_vfs *vfs,
 				   struct l_ino *li, KA_T *vnops));
+# if	defined(HASVXFSRNL)
+_PROTOTYPE(extern int print_vxfs_rnl_path,(struct lfile *lf));
+# endif	/* defined(HASVXFSRNL) */
 #endif	/* defined(HASVXFS) */
 
+#if	defined(HASZONES)
+_PROTOTYPE(extern int enter_zone_arg,(char *zn));
+#endif	/* defined(HASZONES) */
+
 _PROTOTYPE(extern void close_kvm,(void));
 _PROTOTYPE(extern void open_kvm,(void));
 _PROTOTYPE(extern void process_socket,(KA_T sa, char *ty));
