@@ -31,7 +31,7 @@
 
 
 /*
- * $Id: machine.h,v 1.41 2006/03/27 22:57:33 abe Exp $
+ * $Id: machine.h,v 1.43 2008/10/21 16:16:42 abe Exp $
  */
 
 
@@ -76,6 +76,7 @@ typedef struct aio_req { int dummy; } aio_req_t;
 
 # if	solaris>=80000
 #include <sys/wait.h>
+#include <sys/types32.h>
 #define	_KERNEL	1
 #include <netinet/in.h>
 #undef	_KERNEL
@@ -570,6 +571,15 @@ typedef struct aio_req { int dummy; } aio_req_t;
 # endif	/* solaris==20300 || solaris>=20500 */
 
 
+/*
+ * HASTCPUDPSTATE is defined for dialects that have TCP and UDP state
+ * support -- i.e., for the "-stcp|udp:state" option and its associated
+ * speed improvements.
+ */
+
+#define	HASTCPUDPSTATE	1
+
+
 /*
  * HASTMPNODE is defined for those dialects that have tmpnodes.
  */
