@@ -31,7 +31,7 @@
 
 
 /*
- * $Id: lsof.h,v 1.54 2004/12/30 18:39:29 abe Exp $
+ * $Id: lsof.h,v 1.61 2008/10/21 16:21:41 abe Exp $
  */
 
 
@@ -65,9 +65,15 @@
  * Definitions and structures that may be needed by dlsof.h
  */
 
+# if	!defined(INODETYPE)
+#define	INODETYPE	unsigned long	/* node number storage type */
+#define	INODEPSPEC	"l"		/* node number printf specification
+					 * modifier */
+# endif	/* !defined(INODETYPE) */
+
 struct l_dev {
 	dev_t rdev;			/* device */
-	ino_t inode;			/* inode number */
+	INODETYPE inode;		/* inode number */
 	char *name;			/* name */
 	int v;				/* has been verified
 					 * (when DCUnsafe == 1) */
@@ -94,8 +100,10 @@ struct l_dev {
 #define	FF_DEFERIND	"DFI"
 #define	FF_DELAY	"DLY"
 #define	FF_DIRECT	"DIR"
+#define	FF_DIRECTORY	"DTY"
 #define	FF_DOCLONE	"DOCL"
 #define	FF_DSYNC	"DSYN"
+#define	FF_EVTONLY	"EVO"
 #define	FF_EXCL		"EXCL"
 #define	FF_EXEC		"EX"
 #define	FF_EXLOCK	"XL"
@@ -119,9 +127,11 @@ struct l_dev {
 #define	FF_NMFS		"NMFS"
 #define	FF_NDELAY	"ND"
 #define	FF_NET		"NET"
+#define	FF_NOATM	"NATM"
 #define	FF_NOCACHE	"NC"
 #define	FF_NOCTTY	"NTTY"
 #define	FF_NODSYNC	"NDSY"
+#define	FF_NOFOLNK	"NFLK"
 #define	FF_NOTOSTOP	"NOTO"
 #define	FF_NSHARE	"NSH"
 #define	FF_OLRMIRROR	"OLRM"
@@ -129,6 +139,7 @@ struct l_dev {
 #define	FF_POSIX_PIPE	"PP"
 #define	FF_RAIOSIG	"RAIO"
 #define	FF_RCACH	"RC"
+#define	FF_RDWR		"RW"
 #define	FF_READ		"R"
 #define	FF_REVOKED	"REV"
 #define	FF_RSHARE	"RSH"
@@ -169,6 +180,7 @@ struct l_dev {
 #define	POF_INUSE	"USE"
 #define	POF_MAPPED	"MP"
 #define	POF_FSHMAT	"SHMT"
+#define	POF_RESERVED	"OPIP"
 #define	POF_RSVWT	"RSVW"
 
 
@@ -207,7 +219,6 @@ extern int optind;
 #define	CRC_TBLL	256		/* crc table length for software */
 #define	CRC_BITS	8		/* number of bits contributing */
 # endif	/* defined(HASDCACHE) */
-
 #define	CMDL		9		/* maximum number of characters from
 					 * command name to print in COMMAND
 					 * column */
@@ -230,6 +241,11 @@ extern int optind;
 					 * macro */
 # endif	/* !defined(GET_MIN_DEV) */
 
+# if	defined(HASSELINUX)
+#define	HASHCNTX	128		/* security context hash bucket count
+					 * -- MUST BE A POWER OF 2!!! */
+# endif	/* defined(HASSELINUX) */
+
 # if	defined(HASZONES)
 #define	HASHZONE	128		/* zone hash bucket count -- MUST BE
 					 * A POWER OF 2!!! */
@@ -288,32 +304,46 @@ static struct utmp dummy_utmp;		/* to get login name length */
 #define	N_CFS		8		/* CFS node */
 #define	N_CHR		9		/* character device node */
 #define	N_COM		10		/* streams common device node */
-#define	N_DEV		11		/* DEV FS node */
-#define	N_DOOR		12		/* DOOR node */
-#define	N_FD		13		/* FD node */
-#define	N_FIFO		14		/* FIFO node */
-#define	N_HSFS		15		/* High Sierra node */
-#define	N_KERN		16		/* BSD /kern node */
-#define	N_LOFS		17		/* loopback node */
-#define	N_MNT		18		/* mount file system device node */
-#define	N_MPC		19		/* multiplexed device node */
-#define	N_MVFS		20		/* multi-volume file system node (?) */
-#define	N_NFS		21		/* NFS node */
-#define	N_NFS4		22		/* NFS version 4 node */
-#define	N_NM		23		/* named file system node */
-#define	N_OBJF		24		/* objfs file system node */
-#define	N_PCFS		25		/* PC file system node */
-#define	N_PIPE		26		/* pipe device node */
-#define	N_PROC		27		/* /proc node */
-#define	N_PSEU		28		/* pseudofs node */
-#define	N_SAMFS		29		/* Solaris SAM-FS */
-#define	N_SOCK		30		/* sock_vnodeops node */
-#define	N_SPEC		31		/* spec_vnodeops node */
-#define	N_STREAM	32		/* stream node */
-#define	N_TMP		33		/* tmpfs node */
-#define	N_UFS		34		/* UNIX file system node */
-#define	N_VXFS		35		/* Veritas file system node */
-#define	N_XFS		36		/* XFS node */
+#define	N_CTFSADIR	11		/* Solaris CTFS adir node */
+#define	N_CTFSBUND	12		/* Solaris CTFS bundle node */
+#define	N_CTFSCDIR	13		/* Solaris CTFS cdir node */
+#define	N_CTFSCTL	14		/* Solaris CTFS ctl node */
+#define	N_CTFSEVT	15		/* Solaris CTFS events node */
+#define	N_CTFSLATE	16		/* Solaris CTFS latest node */
+#define	N_CTFSROOT	17		/* Solaris CTFS root node */
+#define	N_CTFSSTAT	18		/* Solaris CTFS status node */
+#define	N_CTFSSYM	19		/* Solaris CTFS symbolic node */
+#define	N_CTFSTDIR	20		/* Solaris CTFS type node */
+#define	N_CTFSTMPL	21		/* Solaris CTFS template node */
+#define	N_DEV		22		/* DEV FS node */
+#define	N_DOOR		23		/* DOOR node */
+#define	N_FD		24		/* FD node */
+#define	N_FIFO		25		/* FIFO node */
+#define	N_HSFS		26		/* High Sierra node */
+#define	N_KERN		27		/* BSD /kern node */
+#define	N_LOFS		28		/* loopback node */
+#define	N_MNT		29		/* mount file system device node */
+#define	N_MPC		30		/* multiplexed device node */
+#define	N_MVFS		31		/* multi-volume file system node (?) */
+#define	N_NFS		32		/* NFS node */
+#define	N_NFS4		33		/* NFS version 4 node */
+#define	N_NM		34		/* named file system node */
+#define	N_OBJF		35		/* objfs file system node */
+#define	N_PCFS		36		/* PC file system node */
+#define	N_PIPE		37		/* pipe device node */
+#define	N_PORT		38		/* port node */
+#define	N_PROC		39		/* /proc node */
+#define	N_PSEU		49		/* pseudofs node */
+#define	N_SAMFS		41		/* Solaris SAM-FS */
+#define	N_SANFS		42		/* AIX SANFS */
+#define	N_SOCK		43		/* sock_vnodeops node */
+#define	N_SPEC		44		/* spec_vnodeops node */
+#define	N_STREAM	45		/* stream node */
+#define	N_TMP		46		/* tmpfs node */
+#define	N_UFS		47		/* UNIX file system node */
+#define	N_VXFS		48		/* Veritas file system node */
+#define	N_XFS		49		/* XFS node */
+#define	N_ZFS		50		/* ZFS node */
 
 # if	!defined(OFFDECDIG)
 #define	OFFDECDIG	8		/* maximum number of digits in the
@@ -335,6 +365,8 @@ static struct utmp dummy_utmp;		/* to get login name length */
 #define TCPTPI_WINDOWS	0x0008		/* report TCP/TPI window sizes */
 #define	TCPTPI_ALL	(TCPTPI_QUEUES | TCPTPI_STATE | TCPTPI_WINDOWS)
 					/* report all TCP/TPI info */
+#define	TCPUDPALLOC	32		/* allocation amount for TCP and UDP
+					 * state tables */
 #define	TMLIMIT		15		/* readlink() & stat() timeout sec */
 #define	TMLIMMIN	2		/* minimum timeout */
 #define	TYPEL		8		/* type character length */
@@ -368,6 +400,8 @@ static struct utmp dummy_utmp;		/* to get login name length */
 
 #define	CMDTTL		"COMMAND"
 extern int CmdColW;
+#define	CNTXTTL		"SECURITY-CONTEXT"
+extern int CntxColW;
 #define DEVTTL		"DEVICE"
 extern int DevColW;
 #define	FCTTL		"FCT"
@@ -413,20 +447,22 @@ extern int ZoneColW;
 					 * by PID or UID */
 #define	PS_SEC		2		/* secondary process selection -- e.g.,
 					 * by directory or file */
-#define	SELCMD		0x001		/* select process by command name */
-#define	SELFD		0x002		/* select file by descriptor name */
-#define	SELNA		0x004		/* select socket by address (-i@...) */
-#define	SELNET		0x008		/* select Internet socket files (-i) */
-#define	SELNFS		0x010		/* select NFS files (-N) */
-#define	SELNLINK	0x020		/* select based on link count */
-#define	SELNM		0x040		/* select by name */
-#define	SELPGID		0x080		/* select process group IDs (-g) */
-#define	SELPID		0x100		/* select PIDs (-p) */
-#define	SELUID		0x200		/* select UIDs (-u) */
-#define	SELUNX		0x400		/* select UNIX socket (-U) */
-#define	SELZONE		0x800		/* select zone (-z) */
-#define	SELALL		(SELCMD|SELFD|SELNA|SELNET|SELNM|SELNFS|SELPID|SELUID|SELUNX|SELZONE)
-#define	SELPROC		(SELCMD|SELPGID|SELPID|SELUID|SELZONE)
+#define	SELCMD		0x0001		/* select process by command name */
+#define	SELCNTX		0x0002		/* select security context (-Z) */
+#define	SELFD		0x0004		/* select file by descriptor name */
+#define	SELNA		0x0008		/* select socket by address (-i@...) */
+#define	SELNET		0x0010		/* select Internet socket files (-i) */
+#define	SELNFS		0x0020		/* select NFS files (-N) */
+#define	SELNLINK	0x0040		/* select based on link count */
+#define	SELNM		0x0080		/* select by name */
+#define	SELPGID		0x0100		/* select process group IDs (-g) */
+#define	SELPID		0x0200		/* select PIDs (-p) */
+#define	SELUID		0x0400		/* select UIDs (-u) */
+#define	SELUNX		0x0800		/* select UNIX socket (-U) */
+#define	SELZONE		0x1000		/* select zone (-z) */
+#define	SELEXCLF	0x2000		/* file selection excluded */
+#define	SELALL		(SELCMD|SELCNTX|SELFD|SELNA|SELNET|SELNM|SELNFS|SELPID|SELUID|SELUNX|SELZONE)
+#define	SELPROC		(SELCMD|SELCNTX|SELPGID|SELPID|SELUID|SELZONE)
 					/* process selecters */
 #define	SELFILE		(SELFD|SELNFS|SELNLINK|SELNM)	/* file selecters */
 #define	SELNW		(SELNA|SELNET|SELUNX)		/* network selecters */
@@ -440,7 +476,7 @@ struct afsnode {			/* AFS pseudo-node structure */
 	dev_t dev;
 	unsigned char ino_st;		/* 1 if inode has a value */
 	unsigned char nlink_st;		/* 1 if nlink has a value */
-	unsigned long inode;
+	INODETYPE inode;
 	unsigned long size;
 	long nlink;
 };
@@ -469,7 +505,9 @@ extern struct drive_Nl Drive_Nl[];	/* defined in dstore.c */
 
 struct int_lst {
 	int i;				/* integer argument */
-	int f;				/* find state */
+	int f;				/* find state -- meaningful only if
+					 * x == 0 */
+	int x;				/* excluded state */
 };
 
 typedef struct lsof_rx {		/* regular expression table entry */
@@ -505,11 +543,24 @@ extern int CkPasswd;
 struct str_lst {
 	char *str;			/* string */
 	int len;			/* string length */
-	int f;				/* selected string find state */
+	short f;			/* selected string find state */
+	short x;			/* exclusion (if non-zero) */
 	struct str_lst *next;		/* next list entry */
 };
 extern struct str_lst *Cmdl;
 extern int CmdLim;
+extern int Cmdni;
+extern int Cmdnx;
+ 
+# if	defined(HASSELINUX)
+typedef struct cntxlist {
+	char *cntx;			/* zone name */
+	int f;				/* "find" flag (used only in CntxArg) */
+	struct cntxlist *next;		/* next zone hash entry */
+} cntxlist_t;
+extern cntxlist_t *CntxArg;
+extern int CntxStatus;
+# endif	/* defined(HASSELINUX) */
 
 # if	defined(HASDCACHE)
 extern unsigned DCcksum;
@@ -533,6 +584,7 @@ extern int ErrStat;
 extern uid_t Euid;
 extern int Fand;
 extern int Fblock;
+extern int Fcntx;
 extern int Ffield;
 extern int Ffilesys;
 extern int Fhelp;
@@ -555,6 +607,7 @@ extern int Fpgid;
 extern int Fppid;
 extern int Fsize;
 extern int Fsv;
+extern int FsvByf;
 extern int FsvFlagX;
 extern int Ftcptpi;
 extern int Fterse;
@@ -594,6 +647,8 @@ extern struct fieldsel FieldSel[];
 extern int Hdr;
 
 enum IDType {PGID, PID};
+extern char *InodeFmt_d;
+extern char *InodeFmt_x;
 
 struct lfile {
 	char access;
@@ -615,6 +670,11 @@ struct lfile {
 
 	unsigned char lmi_srch;		/* local mount info search status:
 					 * 1 = printname() search required */
+
+# if	defined(HASMNTSTAT)
+	unsigned char mnt_stat;		/* mount point stat(2) status */
+# endif	/* defined(HASMNTSTAT) */
+
 	unsigned char nlink_def;	/* link count definition status */
 	unsigned char off_def;		/* offset definition status */
 	unsigned char rdev_def;		/* rdev definition status */
@@ -634,14 +694,14 @@ struct lfile {
 	SZOFFTYPE sz;
 	dev_t dev;
 	dev_t rdev;
-	unsigned long inode;
+	INODETYPE inode;
 	long nlink;			/* link count */
 	char *dev_ch;
 	char *fsdir;			/* file system directory */
 	char *fsdev;			/* file system device */
 
 # if	defined(HASFSINO)
-	unsigned long fs_ino;		/* file system inode number */
+	INODETYPE fs_ino;		/* file system inode number */
 # endif	/* defined HASFSINO) */
 
 	struct linaddr {		/* local Internet address information */
@@ -669,13 +729,15 @@ struct lfile {
 	    } state;
 
 # if	defined(HASSOOPT)
-	    unsigned char qlens;	/* qlen status: 0 = none */
-	    unsigned char qlims;	/* qlim status: 0 = none */
-	    unsigned char rbszs;	/* rbsz status: 0 = none */
-	    unsigned char sbszs;	/* sbsz status: 0 = none */
+	    unsigned char pqlens;	/* pqlen status: 0 = none */
+	    unsigned char qlens;	/* qlen status:  0 = none */
+	    unsigned char qlims;	/* qlim status:  0 = none */
+	    unsigned char rbszs;	/* rbsz status:  0 = none */
+	    unsigned char sbszs;	/* sbsz status:  0 = none */
 	    int kai;			/* TCP keep-alive interval */
 	    int ltm;			/* TCP linger time */
 	    unsigned int opt;		/* socket options */
+	    unsigned int pqlen;		/* partial connection queue length */
 	    unsigned int qlen;		/* connection queue length */
 	    unsigned int qlim;		/* connection queue limit */
 	    unsigned long rbsz;		/* receive buffer size */
@@ -741,6 +803,11 @@ extern struct lfile *Lf, *Plf;
 
 struct lproc {
 	char *cmd;			/* command name */
+ 
+# if	defined(HASSELINUX)
+	char *cntx;			/* security context */
+# endif	/* defined(HASSELINUX) */
+
 	short sf;			/* select flags -- SEL* symbols */
 	short pss;			/* state: 0 = not selected
 				 	 *	  1 = wholly selected
@@ -787,7 +854,11 @@ extern long Nlink;
 extern int Nlproc;
 extern char *Nmlst;
 extern int Npgid;
+extern int Npgidi;
+extern int Npgidx;
 extern int Npid;
+extern int Npidi;
+extern int Npidx;
 extern int Npuns;
 extern int Ntype;
 extern int Nuid;
@@ -822,7 +893,7 @@ struct procfsid {
 	unsigned char f;		/* match found if == 1 */
 
 #  if	defined(HASPINODEN)
-	unsigned long inode;		/* search inode number */
+	INODETYPE inode;		/* search inode number */
 #  endif	/* defined(HASPINODEN) */
 
 	struct procfsid *next;		/* forward link */
@@ -849,17 +920,33 @@ extern char *SzOffFmt_0t;
 extern char *SzOffFmt_d;
 extern char *SzOffFmt_dv;
 extern char *SzOffFmt_x;
+extern int TcpStAlloc;
+extern unsigned char *TcpStI;
+extern int TcpStIn;
+extern int TcpStOff;
+extern unsigned char *TcpStX;
+extern int TcpStXn;
+extern int TcpNstates;
+extern char **TcpSt;
 extern char Terminator;
 extern int TmLimit;
+extern int UdpStAlloc;
+extern unsigned char *UdpStI;
+extern int UdpStIn;
+extern int UdpStOff;
+extern unsigned char *UdpStX;
+extern int UdpStXn;
+extern int UdpNstates;
+extern char **UdpSt;
 
-#if	defined(HASZONES)
+# if	defined(HASZONES)
 typedef struct znhash {
 	char *zn;			/* zone name */
 	int f;				/* "find" flag (used only in ZoneArg) */
 	struct znhash *next;		/* next zone hash entry */
 } znhash_t;
 extern znhash_t **ZoneArg;
-#endif	/* defined(HASZONES) */
+# endif	/* defined(HASZONES) */
 
 #include "proto.h"
 #include "dproto.h"
